#!/usr/bin/env python3
import click

from allotropy.parser_factory import update_supported_instruments


@click.command()
def _update_supported_instruments() -> None:
    """Update parser table in supported instruments."""
    update_supported_instruments()


if __name__ == "__main__":
    _update_supported_instruments()



from allotropy.allotrope.models.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.spectrophotometry.benchling._2023._12.spectrophotometry import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.thermo_fisher_genesys30.constants import DISPLAY_NAME
from allotropy.parsers.thermo_fisher_genesys30.thermo_fisher_genesys30_reader import (
    ThermoFisherGenesys30Reader,
)
from allotropy.parsers.thermo_fisher_genesys30.thermo_fisher_genesys30_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.vendor_parser import VendorParser


class ThermoFisherGenesys30Parser(VendorParser[Data, Model]):
    DISPLAY_NAME = DISPLAY_NAME
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = ThermoFisherGenesys30Reader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        reader = ThermoFisherGenesys30Reader(named_file_contents)
        return Data(
            create_metadata(reader.header, named_file_contents.original_file_path),
            create_measurement_groups(reader.header, reader.data),
        )



from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterator
from dataclasses import dataclass
from io import StringIO
import re

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    TransmittedLightSetting,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    ImageFeature,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.exceptions import (
    AllotropeConversionError,
)
from allotropy.parsers.agilent_gen5.agilent_gen5_structure import HeaderData
from allotropy.parsers.agilent_gen5_image.constants import (
    AUTOFOCUS_STRINGS,
    CHANNEL_HEADER_REGEX,
    DEFAULT_SOFTWARE_NAME,
    DETECTION_TYPE,
    DetectionType,
    DETECTOR_DISTANCE_REGEX,
    DEVICE_TYPE,
    METADATA_ONLY,
    ReadType,
    SETTINGS_SECTION_REGEX,
    TRANSMITTED_LIGHT_MAP,
    UNSUPPORTED_READ_TYPE_ERROR,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.lines_reader import LinesReader, SectionLinesReader
from allotropy.parsers.utils.pandas import read_csv
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    try_float,
    try_float_or_nan,
    try_float_or_none,
)


def parse_settings(settings: list[str]) -> tuple[dict[str, str], list[str]]:
    """Returns a dictionary containing all key values identified in a list of settings.

    If there are additional (non key value) settings, they are also returned under the
    'non_kv_settings' key.

    Supported settings lines format:
        - key: value -> returned as {'key': 'value'}
        - key1: value1, key2: value2, ... -> returned as {'key1': 'value1', key2', 'value2', ...}
        - Non keyvalue setting -> returned as {'non_kv_settings': ['Non keyvalue setting', ...]}
    """
    non_kv_settings: list[str] = []
    settings_dict: dict[str, str] = {}
    for line in settings:
        strp_line = str(line.strip())

        line_data: list[str] = strp_line.split(", ")
        for read_datum in line_data:
            splitted_datum = read_datum.split(": ")
            if len(splitted_datum) == 1:
                non_kv_settings.append(splitted_datum[0])
            elif len(splitted_datum) == 2:
                settings_dict[splitted_datum[0]] = splitted_datum[1]

    return settings_dict, non_kv_settings


@dataclass
class InstrumentSettings:
    auto_focus: bool
    detector_distance: float | None
    fluorescent_tag: str | None = None
    excitation_wavelength: float | None = None
    detector_wavelength: float | None = None
    transmitted_light: TransmittedLightSetting | None = None
    illumination: float | None = None
    exposure_duration: float | None = None
    detector_gain: str | None = None

    @classmethod
    def create(cls, settings_lines: list[str]) -> InstrumentSettings:
        fluorescence_channel_settings = cls._get_fluorescence_settings(
            settings_lines[0]
        )
        transmitted_light = None
        if not fluorescence_channel_settings:
            transmitted_light = cls._get_transmitted_light_out_of_channel_settings(
                settings_lines[0]
            )
        settings_dict, non_kv_settings = parse_settings(settings_lines[1:])
        if not transmitted_light:
            transmitted_light = cls._get_transmitted_light(non_kv_settings)

        if exposure_duration := settings_dict.get("Integration time"):
            exposure_duration = str(exposure_duration).split()[0]

        return InstrumentSettings(
            auto_focus=cls._get_auto_focus(non_kv_settings),
            detector_distance=cls._get_detector_distance(non_kv_settings),
            fluorescent_tag=fluorescence_channel_settings.get("fluorescent_tag"),
            excitation_wavelength=try_float_or_none(
                fluorescence_channel_settings.get("excitation_wavelength")
            ),
            detector_wavelength=try_float_or_none(
                fluorescence_channel_settings.get("detector_wavelength")
            ),
            transmitted_light=transmitted_light,
            illumination=try_float_or_none(settings_dict.get("LED intensity")),
            exposure_duration=try_float_or_none(exposure_duration),
            detector_gain=settings_dict.get("Camera gain"),
        )

    @classmethod
    def _get_fluorescence_settings(cls, settings_header: str) -> dict[str, str]:
        if matches := re.match(CHANNEL_HEADER_REGEX, settings_header):
            return {key: str(value) for key, value in matches.groupdict().items()}
        return {}

    @classmethod
    def _get_auto_focus(cls, settings: list[str]) -> bool:
        return any(str_ in settings for str_ in AUTOFOCUS_STRINGS)

    @classmethod
    def _get_detector_distance(cls, settings: list[str]) -> float | None:
        if match := re.search(DETECTOR_DISTANCE_REGEX, "\n".join(settings)):
            return try_float(match.groups()[0], "Detector Distance")
        return None

    @classmethod
    def _get_transmitted_light(
        cls, settings: list[str]
    ) -> TransmittedLightSetting | None:
        for line in settings:
            if line in TRANSMITTED_LIGHT_MAP:
                return TRANSMITTED_LIGHT_MAP[line]
        return None

    @classmethod
    def _get_transmitted_light_out_of_channel_settings(
        cls, channel_setting: str
    ) -> TransmittedLightSetting | None:
        keys = "|".join(TRANSMITTED_LIGHT_MAP.keys())
        pattern = rf"\b(?:{keys})\b"
        if match := re.search(pattern, channel_setting):
            return TRANSMITTED_LIGHT_MAP[match.group()]
        return None


@dataclass(frozen=True)
class ReadSection:
    image_mode: DetectionType
    magnification_setting: float | None
    image_count_setting: float | None
    instrument_settings_list: list[InstrumentSettings]

    @classmethod
    def create(cls, reader: LinesReader) -> ReadSection:
        top_read_lines = list(reader.pop_until(SETTINGS_SECTION_REGEX))

        detection_type = cls._get_detection_type("\n".join(top_read_lines))
        instrument_settings_list = cls._get_instrument_settings_list(reader)

        bottom_read_lines = list(reader.pop_until_empty())
        read_settings, _ = parse_settings(top_read_lines + bottom_read_lines)

        if objective := read_settings.get("Objective"):
            objective = str(objective).replace("x", "")

        return ReadSection(
            image_mode=detection_type,
            magnification_setting=try_float_or_none(objective),
            image_count_setting=cls._get_image_count_setting(read_settings),
            instrument_settings_list=instrument_settings_list,
        )

    @classmethod
    def _get_detection_type(cls, read_chunk: str) -> DetectionType:
        if DetectionType.SINGLE_IMAGE.value in read_chunk:
            return DetectionType.SINGLE_IMAGE
        elif DetectionType.MONTAGE.value in read_chunk:
            return DetectionType.MONTAGE
        elif DetectionType.Z_STACK.value in read_chunk:
            return DetectionType.Z_STACK

        msg = f"Measurement mode not found; expected to find one of {sorted(DetectionType._member_names_)}."

        raise AllotropeConversionError(msg)

    @classmethod
    def _get_settings_sections(cls, reader: LinesReader) -> Iterator[list[str]]:
        while True:
            initial_line = reader.get() or ""
            if not re.search(SETTINGS_SECTION_REGEX, initial_line):
                break
            reader.pop()
            lines = [initial_line, *reader.pop_until(r"^\t\w")]
            yield lines

    @classmethod
    def _get_instrument_settings_list(
        cls, reader: LinesReader
    ) -> list[InstrumentSettings]:
        settings_list = []
        auto_focus = False
        for idx, settings_section in enumerate(cls._get_settings_sections(reader)):
            instrument_settings = InstrumentSettings.create(settings_section)
            # The autofocus setting is only reported in the first section but applies to all others
            if idx == 0:
                auto_focus = instrument_settings.auto_focus
            else:
                instrument_settings.auto_focus = auto_focus
            settings_list.append(instrument_settings)
        return settings_list

    @classmethod
    def _get_image_count_setting(cls, read_settings: dict[str, str]) -> float | None:
        montage_rows = read_settings.get("Montage rows")
        montage_columns = read_settings.get("columns")
        if montage_rows and montage_columns:
            return try_float(montage_rows, "Montage rows") * try_float(
                montage_columns, "Montage columns"
            )
        return None


@dataclass(frozen=True)
class ReadData:
    read_sections: list[ReadSection]

    @classmethod
    def create(cls, procedure_details: list[str]) -> ReadData:
        read_type = cls._get_read_type("\n".join(procedure_details))
        if read_type != ReadType.IMAGE:
            raise AllotropeConversionError(UNSUPPORTED_READ_TYPE_ERROR)

        section_lines_reader = SectionLinesReader(procedure_details)

        return ReadData(
            read_sections=[
                ReadSection.create(read_section)
                for read_section in section_lines_reader.iter_sections("^Read\t")
            ]
        )

    @classmethod
    def _get_read_type(cls, procedure_details: str) -> ReadType:
        if ReadType.KINETIC.value in procedure_details:
            return ReadType.KINETIC
        elif ReadType.AREASCAN.value in procedure_details:
            return ReadType.AREASCAN
        elif ReadType.SPECTRAL.value in procedure_details:
            return ReadType.SPECTRAL
        elif ReadType.ENDPOINT.value in procedure_details:
            return ReadType.ENDPOINT
        elif ReadType.IMAGE.value in procedure_details:
            return ReadType.IMAGE

        msg = f"Read type not found; expected to find one of {sorted(ReadType._member_names_)}."
        raise AllotropeConversionError(msg)


def create_results(
    result_lines: list[str] | None,
    header_data: HeaderData,
    read_data: ReadData,
    sample_identifiers: dict[str, str],
) -> list[MeasurementGroup]:
    if not result_lines:
        return [
            MeasurementGroup(
                measurement_time=header_data.datetime,
                plate_well_count=header_data.plate_well_count,
                analytical_method_identifier=header_data.protocol_file_path,
                experimental_data_identifier=header_data.experiment_file_path,
                measurements=[
                    _create_measurement(
                        NOT_APPLICABLE,
                        header_data,
                        read_section,
                        instrument_settings,
                        METADATA_ONLY,
                        None,
                    )
                    for read_section in read_data.read_sections
                    for instrument_settings in read_section.instrument_settings_list
                ],
            )
        ]
    if result_lines[0].strip() != "Results":
        msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
        raise AllotropeConversionError(msg)

    # Create dataframe from tabular data and forward fill empty values in index
    data = read_csv(StringIO("\n".join(result_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis="index").values)

    image_features = defaultdict(list)

    for row_name, row in data.iterrows():
        feature_name = row.iloc[-1]
        for col_index, value in enumerate(row.iloc[:-1]):
            well_pos = f"{row_name}{col_index + 1}"
            well_value = try_float_or_nan(value)

            image_features[well_pos].append(
                ImageFeature(
                    identifier=random_uuid_str(),
                    feature=feature_name,
                    result=well_value,
                )
            )

    groups = []
    num_measurements = sum(
        len(read_section.instrument_settings_list)
        for read_section in read_data.read_sections
    )
    for well_position in image_features:
        processed_data = ProcessedData(
            identifier=random_uuid_str(), features=image_features[well_position]
        )
        measurements = [
            _create_measurement(
                well_position,
                header_data,
                read_section,
                instrument_settings,
                sample_identifiers.get(well_position),
                processed_data if num_measurements == 1 else None,
            )
            for read_section in read_data.read_sections
            for instrument_settings in read_section.instrument_settings_list
        ]

        groups.append(
            MeasurementGroup(
                measurement_time=header_data.datetime,
                plate_well_count=len(image_features),
                analytical_method_identifier=header_data.protocol_file_path,
                experimental_data_identifier=header_data.experiment_file_path,
                measurements=measurements,
                processed_data=processed_data if num_measurements > 1 else None,
            )
        )

    return groups


def create_metadata(header_data: HeaderData) -> Metadata:
    return Metadata(
        device_identifier=NOT_APPLICABLE,
        model_number=header_data.model_number or NOT_APPLICABLE,
        equipment_serial_number=header_data.equipment_serial_number,
        software_name=DEFAULT_SOFTWARE_NAME,
        software_version=header_data.software_version,
        file_name=header_data.file_name,
        unc_path=header_data.unc_path,
    )


def _create_measurement(
    well_position: str,
    header_data: HeaderData,
    read_section: ReadSection,
    instrument_settings: InstrumentSettings,
    sample_identifier: str | None,
    processed_data: ProcessedData | None,
) -> Measurement:
    return Measurement(
        type_=MeasurementType.OPTICAL_IMAGING,
        device_type=DEVICE_TYPE,
        identifier=random_uuid_str(),
        sample_identifier=sample_identifier
        or f"{header_data.well_plate_identifier} {well_position}",
        location_identifier=well_position,
        well_plate_identifier=header_data.well_plate_identifier,
        detection_type=DETECTION_TYPE,
        detector_wavelength_setting=instrument_settings.detector_wavelength,
        excitation_wavelength_setting=instrument_settings.excitation_wavelength,
        # TODO: this setting won't get reported at the moment since Gen5 only reports it
        # in micrometers and we don't do conversions on the adapters at the moment
        # detector_distance_setting=instrument_settings.detector_distance,
        detector_gain_setting=instrument_settings.detector_gain,
        magnification_setting=read_section.magnification_setting,
        illumination_setting=instrument_settings.illumination,
        illumination_setting_unit=UNITLESS,
        transmitted_light_setting=instrument_settings.transmitted_light,
        auto_focus_setting=instrument_settings.auto_focus,
        image_count_setting=read_section.image_count_setting,
        fluorescent_tag_setting=instrument_settings.fluorescent_tag,
        exposure_duration_setting=instrument_settings.exposure_duration,
        processed_data=processed_data,
    )



import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import determine_encoding
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    read_csv,
    read_excel,
)


class RevvityMatrixReader:
    SUPPORTED_EXTENSIONS = "csv,xlsx"
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        if named_file_contents.extension == "csv":
            contents = named_file_contents.contents.read()
            encoding = (
                determine_encoding(contents, named_file_contents.encoding)
                if isinstance(contents, bytes)
                else None
            )
            named_file_contents.contents.seek(0)
            df = read_csv(named_file_contents.contents, encoding=encoding)
        else:
            df = read_excel(named_file_contents.contents)
            # Reading a percent value (50%) in read_excel results in a decimal: 0.5
            # Detect and adjust value back to 0-100%
            first_row = df_to_series_data(df, 0)
            viability = first_row[str, "Viability"]
            if "%" not in first_row[str, "Viability"] and float(viability) < 1:
                df["Viability"] = df["Viability"] * 100
        self.data = df



from collections import defaultdict
from datetime import tzinfo
from enum import Enum
from pathlib import Path
from typing import Any

from allotropy.allotrope.schema_parser.path_util import ROOT_DIR
from allotropy.parsers.agilent_gen5.agilent_gen5_parser import AgilentGen5Parser
from allotropy.parsers.agilent_gen5_image.agilent_gen5_image_parser import (
    AgilentGen5ImageParser,
)
from allotropy.parsers.agilent_tapestation_analysis.agilent_tapestation_analysis_parser import (
    AgilentTapestationAnalysisParser,
)
from allotropy.parsers.appbio_absolute_q.appbio_absolute_q_parser import (
    AppbioAbsoluteQParser,
)
from allotropy.parsers.appbio_quantstudio.appbio_quantstudio_parser import (
    AppBioQuantStudioParser,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_parser import (
    AppBioQuantStudioDesignandanalysisParser,
)
from allotropy.parsers.beckman_coulter_biomek.beckman_coulter_biomek_parser import (
    BeckmanCoulterBiomekParser,
)
from allotropy.parsers.beckman_pharmspec.beckman_pharmspec_parser import PharmSpecParser
from allotropy.parsers.beckman_vi_cell_blu.vi_cell_blu_parser import ViCellBluParser
from allotropy.parsers.beckman_vi_cell_xr.vi_cell_xr_parser import ViCellXRParser
from allotropy.parsers.benchling_empower.benchling_empower_parser import (
    BenchlingEmpowerParser,
)
from allotropy.parsers.biorad_bioplex_manager.biorad_bioplex_manager_parser import (
    BioradBioplexParser,
)
from allotropy.parsers.bmg_mars.bmg_mars_parser import BmgMarsParser
from allotropy.parsers.cfxmaestro.cfxmaestro_parser import CfxmaestroParser
from allotropy.parsers.chemometec_nc_view.chemometec_nc_view_parser import (
    ChemometecNcViewParser,
)
from allotropy.parsers.chemometec_nucleoview.nucleoview_parser import (
    ChemometecNucleoviewParser,
)
from allotropy.parsers.ctl_immunospot.ctl_immunospot_parser import CtlImmunospotParser
from allotropy.parsers.cytiva_unicorn.cytiva_unicorn_parser import CytivaUnicornParser
from allotropy.parsers.example_weyland_yutani.example_weyland_yutani_parser import (
    ExampleWeylandYutaniParser,
)
from allotropy.parsers.luminex_xponent.luminex_xponent_parser import (
    LuminexXponentParser,
)
from allotropy.parsers.mabtech_apex.mabtech_apex_parser import MabtechApexParser
from allotropy.parsers.methodical_mind.methodical_mind_parser import (
    MethodicalMindParser,
)
from allotropy.parsers.moldev_softmax_pro.softmax_pro_parser import SoftmaxproParser
from allotropy.parsers.msd_workbench.msd_workbench_parser import MSDWorkbenchParser
from allotropy.parsers.novabio_flex2.novabio_flex2_parser import NovaBioFlexParser
from allotropy.parsers.perkin_elmer_envision.perkin_elmer_envision_parser import (
    PerkinElmerEnvisionParser,
)
from allotropy.parsers.qiacuity_dpcr.qiacuity_dpcr_parser import QiacuitydPCRParser
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.revvity_kaleido.kaleido_parser import KaleidoParser
from allotropy.parsers.revvity_matrix.revvity_matrix_parser import (
    RevvityMatrixParser,
)
from allotropy.parsers.roche_cedex_bioht.roche_cedex_bioht_parser import (
    RocheCedexBiohtParser,
)
from allotropy.parsers.roche_cedex_hires.roche_cedex_hires_parser import (
    RocheCedexHiResParser,
)
from allotropy.parsers.tecan_magellan.tecan_magellan_parser import TecanMagellanParser
from allotropy.parsers.thermo_fisher_genesys30.thermo_fisher_genesys30_parser import (
    ThermoFisherGenesys30Parser,
)
from allotropy.parsers.thermo_fisher_genesys_on_board.thermo_fisher_genesys_on_board_parser import (
    ThermoFisherGenesysOnBoardParser,
)
from allotropy.parsers.thermo_fisher_nanodrop_8000.nanodrop_8000_parser import (
    Nanodrop8000Parser,
)
from allotropy.parsers.thermo_fisher_nanodrop_eight.nanodrop_eight_parser import (
    NanodropEightParser,
)
from allotropy.parsers.thermo_fisher_nanodrop_one.thermo_fisher_nanodrop_one_parser import (
    ThermoFisherNanodropOneParser,
)
from allotropy.parsers.thermo_fisher_qubit4.thermo_fisher_qubit4_parser import (
    ThermoFisherQubit4Parser,
)
from allotropy.parsers.thermo_fisher_qubit_flex.thermo_fisher_qubit_flex_parser import (
    ThermoFisherQubitFlexParser,
)
from allotropy.parsers.thermo_fisher_visionlite.thermo_fisher_visionlite_parser import (
    ThermoFisherVisionliteParser,
)
from allotropy.parsers.thermo_skanit.thermo_skanit_parser import ThermoSkanItParser
from allotropy.parsers.unchained_labs_lunatic.unchained_labs_lunatic_parser import (
    UnchainedLabsLunaticParser,
)
from allotropy.parsers.utils.timestamp_parser import TimestampParser
from allotropy.parsers.vendor_parser import VendorParser


class Vendor(Enum):
    AGILENT_GEN5 = "AGILENT_GEN5"
    AGILENT_GEN5_IMAGE = "AGILENT_GEN5_IMAGE"
    AGILENT_TAPESTATION_ANALYSIS = "AGILENT_TAPESTATION_ANALYSIS"
    APPBIO_ABSOLUTE_Q = "APPBIO_ABSOLUTE_Q"
    APPBIO_QUANTSTUDIO = "APPBIO_QUANTSTUDIO"
    APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS = "APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS"
    BENCHLING_EMPOWER = "BENCHLING_EMPOWER"
    BECKMAN_COULTER_BIOMEK = "BECKMAN_COULTER_BIOMEK"
    BMG_MARS = "BMG_MARS"
    BECKMAN_PHARMSPEC = "BECKMAN_PHARMSPEC"
    BECKMAN_VI_CELL_BLU = "BECKMAN_VI_CELL_BLU"
    BECKMAN_VI_CELL_XR = "BECKMAN_VI_CELL_XR"
    BIORAD_BIOPLEX = "BIORAD_BIOPLEX"
    CFXMAESTRO = "CFXMAESTRO"
    CHEMOMETEC_NC_VIEW = "CHEMOMETEC_NC_VIEW"
    CHEMOMETEC_NUCLEOVIEW = "CHEMOMETEC_NUCLEOVIEW"
    CTL_IMMUNOSPOT = "CTL_IMMUNOSPOT"
    CYTIVA_UNICORN = "CYTIVA_UNICORN"
    EXAMPLE_WEYLAND_YUTANI = "EXAMPLE_WEYLAND_YUTANI"
    LUMINEX_XPONENT = "LUMINEX_XPONENT"
    MABTECH_APEX = "MABTECH_APEX"
    METHODICAL_MIND = "METHODICAL_MIND"
    MOLDEV_SOFTMAX_PRO = "MOLDEV_SOFTMAX_PRO"
    MSD_WORKBENCH = "MSD_WORKBENCH"
    REVVITY_MATRIX = "REVVITY_MATRIX"
    NOVABIO_FLEX2 = "NOVABIO_FLEX2"
    PERKIN_ELMER_ENVISION = "PERKIN_ELMER_ENVISION"
    QIACUITY_DPCR = "QIACUITY_DPCR"
    REVVITY_KALEIDO = "REVVITY_KALEIDO"
    ROCHE_CEDEX_BIOHT = "ROCHE_CEDEX_BIOHT"
    ROCHE_CEDEX_HIRES = "ROCHE_CEDEX_HIRES"
    TECAN_MAGELLAN = "TECAN_MAGELLAN"
    THERMO_FISHER_GENESYS30 = "THERMO_FISHER_GENESYS30"
    THERMO_FISHER_GENESYS_ON_BOARD = "THERMO_FISHER_GENESYS_ON_BOARD"
    THERMO_FISHER_NANODROP_8000 = "THERMO_FISHER_NANODROP_8000"
    THERMO_FISHER_NANODROP_EIGHT = "THERMO_FISHER_NANODROP_EIGHT"
    THERMO_FISHER_NANODROP_ONE = "THERMO_FISHER_NANODROP_ONE"
    THERMO_FISHER_QUBIT4 = "THERMO_FISHER_QUBIT4"
    THERMO_FISHER_QUBIT_FLEX = "THERMO_FISHER_QUBIT_FLEX"
    THERMO_FISHER_VISIONLITE = "THERMO_FISHER_VISIONLITE"
    THERMO_SKANIT = "THERMO_SKANIT"
    UNCHAINED_LABS_LUNATIC = "UNCHAINED_LABS_LUNATIC"

    @property
    def display_name(self) -> str:
        return self.get_parser().DISPLAY_NAME

    @property
    def release_state(self) -> ReleaseState:
        return self.get_parser().RELEASE_STATE

    @property
    def supported_extensions(self) -> list[str]:
        return [
            ext.strip() for ext in self.get_parser().SUPPORTED_EXTENSIONS.split(",")
        ]

    @property
    def asm_versions(self) -> list[str]:
        # NOTE: this is a list because soon parsers will support multiple schemas as they are upgraded.
        manifests = [
            Path(manifest) for manifest in [self.get_parser()._get_mapper().MANIFEST]
        ]
        return ["/".join(manifest.parts[-4:-1]).split(".")[0] for manifest in manifests]

    @property
    def technique(self) -> str:
        techniques = [
            Path(manifest).stem
            for manifest in [self.get_parser()._get_mapper().MANIFEST]
        ]
        if not all(tech == techniques[0] for tech in techniques):
            msg = f"Parser {self} supports multiple technique types, if this is expected please update logic."
            raise AssertionError(msg)
        technique = techniques[0].replace("-", " ").title()
        return {
            "Dpcr": "dPCR",
            "Qpcr": "qPCR",
        }.get(technique, technique)

    def get_parser(
        self, default_timezone: tzinfo | None = None
    ) -> VendorParser[Any, Any]:
        timestamp_parser = TimestampParser(default_timezone)
        return _VENDOR_TO_PARSER[self](timestamp_parser)


_VENDOR_TO_PARSER: dict[Vendor, type[VendorParser[Any, Any]]] = {
    Vendor.AGILENT_GEN5: AgilentGen5Parser,
    Vendor.AGILENT_GEN5_IMAGE: AgilentGen5ImageParser,
    Vendor.AGILENT_TAPESTATION_ANALYSIS: AgilentTapestationAnalysisParser,
    Vendor.APPBIO_ABSOLUTE_Q: AppbioAbsoluteQParser,
    Vendor.APPBIO_QUANTSTUDIO: AppBioQuantStudioParser,
    Vendor.APPBIO_QUANTSTUDIO_DESIGNANDANALYSIS: AppBioQuantStudioDesignandanalysisParser,
    Vendor.BECKMAN_COULTER_BIOMEK: BeckmanCoulterBiomekParser,
    Vendor.BECKMAN_PHARMSPEC: PharmSpecParser,
    Vendor.BECKMAN_VI_CELL_BLU: ViCellBluParser,
    Vendor.BECKMAN_VI_CELL_XR: ViCellXRParser,
    Vendor.BENCHLING_EMPOWER: BenchlingEmpowerParser,
    Vendor.BIORAD_BIOPLEX: BioradBioplexParser,
    Vendor.BMG_MARS: BmgMarsParser,
    Vendor.CFXMAESTRO: CfxmaestroParser,
    Vendor.CHEMOMETEC_NC_VIEW: ChemometecNcViewParser,
    Vendor.CHEMOMETEC_NUCLEOVIEW: ChemometecNucleoviewParser,
    Vendor.CTL_IMMUNOSPOT: CtlImmunospotParser,
    Vendor.CYTIVA_UNICORN: CytivaUnicornParser,
    Vendor.EXAMPLE_WEYLAND_YUTANI: ExampleWeylandYutaniParser,
    Vendor.LUMINEX_XPONENT: LuminexXponentParser,
    Vendor.MABTECH_APEX: MabtechApexParser,
    Vendor.METHODICAL_MIND: MethodicalMindParser,
    Vendor.MOLDEV_SOFTMAX_PRO: SoftmaxproParser,
    Vendor.MSD_WORKBENCH: MSDWorkbenchParser,
    Vendor.REVVITY_MATRIX: RevvityMatrixParser,
    Vendor.NOVABIO_FLEX2: NovaBioFlexParser,
    Vendor.PERKIN_ELMER_ENVISION: PerkinElmerEnvisionParser,
    Vendor.QIACUITY_DPCR: QiacuitydPCRParser,
    Vendor.REVVITY_KALEIDO: KaleidoParser,
    Vendor.ROCHE_CEDEX_BIOHT: RocheCedexBiohtParser,
    Vendor.ROCHE_CEDEX_HIRES: RocheCedexHiResParser,
    Vendor.TECAN_MAGELLAN: TecanMagellanParser,
    Vendor.THERMO_FISHER_GENESYS30: ThermoFisherGenesys30Parser,
    Vendor.THERMO_FISHER_GENESYS_ON_BOARD: ThermoFisherGenesysOnBoardParser,
    Vendor.THERMO_FISHER_NANODROP_8000: Nanodrop8000Parser,
    Vendor.THERMO_FISHER_NANODROP_EIGHT: NanodropEightParser,
    Vendor.THERMO_FISHER_NANODROP_ONE: ThermoFisherNanodropOneParser,
    Vendor.THERMO_FISHER_QUBIT4: ThermoFisherQubit4Parser,
    Vendor.THERMO_FISHER_QUBIT_FLEX: ThermoFisherQubitFlexParser,
    Vendor.THERMO_FISHER_VISIONLITE: ThermoFisherVisionliteParser,
    Vendor.THERMO_SKANIT: ThermoSkanItParser,
    Vendor.UNCHAINED_LABS_LUNATIC: UnchainedLabsLunaticParser,
}


def get_table_contents() -> str:
    contents = """The parsers follow maturation levels of: Recommended, Candidate Release, Working Draft.

* Recommended - "General Availability" - the parser has good coverage of input formats, and exports the majority of data from source files.
** Note that while we judge release status based on observed use cases, it is still a judgement call - it is still possible that we miss some cases.
** If you have an example case for an instrument that is not supported or discover a bug, please https://github.com/Benchling-Open-Source/allotropy/issues[open an issue] with sample data and will work to add support!
* Candidate Release - "Limited Availability" - the parser exports correct data for supported cases, but may be missing some functionality, including:
** May not handle all possible input formats from the target instrument software (because they have not been seen before).
** May not export all available data from the input, either because it cannot be supported yet by ASM, or because we have not determined how to add it.
** Increased likelihood of bugs due to lack of "bake time" for discovering issues.
* Working Draft - "Beta" - in development, not ready for production use:
** May be missing enough sample data for us to be confident about correctness of parser
** May be in-progress, with signifcant known TODO work

'''
"""
    table_data: defaultdict[str, list[Vendor]] = defaultdict(list)

    for vendor in Vendor:
        if "example" in str(vendor).lower():
            continue
        table_data[vendor.technique].append(vendor)

    contents += '[cols="4*^.^"]\n'
    contents += "|===\n"
    contents += (
        "|Instrument Category|Instrument Software|Release Status|Exported ASM Schema\n"
    )
    for technique in sorted(table_data):
        vendors = table_data[technique]
        contents += f".{len(vendors)}+|{technique}|{vendors[0].display_name}|{vendors[0].release_state}|{vendors[0].asm_versions[0]}\n"
        for vendor in vendors[1:]:
            contents += f"|{vendor.display_name}|{vendor.release_state}|{vendor.asm_versions[0]}\n"
    contents += "|==="

    return contents


def update_supported_instruments() -> None:
    with open(Path(ROOT_DIR, "SUPPORTED_INSTRUMENT_SOFTWARE.adoc"), "w") as f:
        f.write(get_table_contents())



from xml.etree import ElementTree as ET  # noqa: N817

from allotropy.allotrope.models.adm.electrophoresis.benchling._2024._09.electrophoresis import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.electrophoresis.benchling._2024._09.electrophoresis import (
    Data,
    Mapper,
)
from allotropy.exceptions import AllotropeParsingError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.agilent_tapestation_analysis.agilent_tapestation_analysis_structure import (
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AgilentTapestationAnalysisParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "Agilent TapeStation Analysis"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "xml"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        try:
            root_element = ET.parse(  # noqa: S314
                named_file_contents.contents
            ).getroot()
        except ET.ParseError as e:
            msg = f"There was an error when trying to read the xml file: {e}"
            raise AllotropeParsingError(msg) from e

        measurement_groups, calculated_data = create_measurement_groups(root_element)
        return Data(
            create_metadata(root_element, named_file_contents.original_file_path),
            measurement_groups,
            # NOTE: in current implementation, calculated data is reported at global level for some reason.
            # TODO(nstender): should we move this inside of measurements?
            calculated_data,
        )



DETECTION_TYPE = "optical-imaging"
SOFTWARE_NAME = "Apex"
DEVICE_TYPE = "imager"

IMAGE_FEATURES = [
    "Spot Forming Units (SFU)",
    "Sum of Spot Volume (RSV)",
    "Average Relative Spot Volume (RSV)",
    "Average Relative Spot Volume {filter} (RSV{led_number})",
    "Sum Relative Spot Volume {filter} (RSV{led_number})",
]



from datetime import tzinfo
from typing import Any

from allotropy.allotrope.allotrope import serialize_and_validate_allotrope
from allotropy.exceptions import AllotropeConversionError
from allotropy.named_file_contents import NamedFileContents
from allotropy.parser_factory import Vendor
from allotropy.types import IOType

VendorType = Vendor | str


def allotrope_from_io(
    contents: IOType,
    filepath: str,
    vendor_type: VendorType,
    default_timezone: tzinfo | None = None,
    encoding: str | None = None,
) -> dict[str, Any]:
    model = allotrope_model_from_io(
        contents, filepath, vendor_type, default_timezone, encoding
    )
    return serialize_and_validate_allotrope(model)


def allotrope_model_from_io(
    contents: IOType,
    filepath: str,
    vendor_type: VendorType,
    default_timezone: tzinfo | None = None,
    encoding: str | None = None,
) -> Any:
    try:
        vendor = Vendor(vendor_type)
    except ValueError as e:
        msg = f"Failed to create parser, unregistered vendor: {vendor_type}."
        raise AllotropeConversionError(msg) from e
    named_file_contents = NamedFileContents(contents, filepath, encoding)
    if named_file_contents.extension not in vendor.supported_extensions:
        msg = f"Unsupported file extension '{named_file_contents.extension}' for parser '{vendor.display_name}', expected one of '{vendor.supported_extensions}'."
        raise AllotropeConversionError(msg)
    parser = vendor.get_parser(default_timezone=default_timezone)
    return parser.to_allotrope(named_file_contents)


def allotrope_from_file(
    filepath: str,
    vendor_type: VendorType,
    default_timezone: tzinfo | None = None,
    encoding: str | None = None,
) -> dict[str, Any]:
    model = allotrope_model_from_file(filepath, vendor_type, default_timezone, encoding)
    return serialize_and_validate_allotrope(model)


def allotrope_model_from_file(
    filepath: str,
    vendor_type: VendorType,
    default_timezone: tzinfo | None = None,
    encoding: str | None = None,
) -> Any:
    try:
        with open(filepath, "rb") as f:
            return allotrope_model_from_io(
                f,
                filepath,
                vendor_type,
                default_timezone=default_timezone,
                encoding=encoding,
            )
    except FileNotFoundError as e:
        msg = f"File not found: {filepath}."
        raise AllotropeConversionError(msg) from e



from __future__ import annotations

from pathlib import Path

from allotropy.allotrope.schema_mappers.adm.cell_counting.rec._2024._09.cell_counting import (
    Measurement,
    MeasurementGroup,
    Metadata,
)
from allotropy.parsers.beckman_vi_cell_blu.constants import (
    DEFAULT_ANALYST,
    DEFAULT_MODEL_NUMBER,
    DETECTION_TYPE,
    DEVICE_TYPE,
    VICELL_BLU_SOFTWARE_NAME,
)
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


def create_measurement_group(data: SeriesData) -> MeasurementGroup:
    total_cell_count = data.get(float, "Cell count", validate=SeriesData.NOT_NAN)
    total_cell_count = (
        total_cell_count if total_cell_count is None else round(total_cell_count)
    )
    viable_cell_count = data.get(float, "Viable cells", validate=SeriesData.NOT_NAN)
    viable_cell_count = (
        viable_cell_count if viable_cell_count is None else round(viable_cell_count)
    )

    return MeasurementGroup(
        analyst=data.get(str, "Analysis by", DEFAULT_ANALYST),
        measurements=[
            Measurement(
                measurement_identifier=random_uuid_str(),
                timestamp=data[str, "Analysis date/time"],
                sample_identifier=data[str, "Sample ID"],
                cell_type_processing_method=data.get(str, "Cell type"),
                minimum_cell_diameter_setting=data.get(float, "Minimum Diameter (μm)"),
                maximum_cell_diameter_setting=data.get(float, "Maximum Diameter (μm)"),
                cell_density_dilution_factor=data.get(float, "Dilution"),
                viability=data[float, "Viability (%)"],
                viable_cell_density=data[float, "Viable (x10^6) cells/mL"],
                total_cell_count=total_cell_count,
                total_cell_density=data.get(float, "Total (x10^6) cells/mL"),
                average_total_cell_diameter=data.get(float, "Average diameter (μm)"),
                average_live_cell_diameter=data.get(
                    float, "Average viable diameter (μm)"
                ),
                viable_cell_count=viable_cell_count,
                average_total_cell_circularity=data.get(float, "Average circularity"),
                average_viable_cell_circularity=data.get(
                    float, "Average viable circularity"
                ),
            )
        ],
    )


def create_metadata(file_path: str) -> Metadata:
    path = Path(file_path)
    return Metadata(
        asm_file_identifier=path.with_suffix(".json").name,
        data_system_instance_id=NOT_APPLICABLE,
        device_identifier=NOT_APPLICABLE,
        device_type=DEVICE_TYPE,
        detection_type=DETECTION_TYPE,
        model_number=DEFAULT_MODEL_NUMBER,
        software_name=VICELL_BLU_SOFTWARE_NAME,
        file_name=Path(file_path).name,
        unc_path=file_path,
    )



import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.pandas import SeriesData
from allotropy.parsers.utils.values import assert_not_none


class BeckmanCoulterBiomekReader:
    SUPPORTED_EXTENSIONS = "csv"
    header: SeriesData
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        reader = CsvReader.create(named_file_contents)

        header_lines = list(reader.pop_until(".*Well Index.*"))
        header_dict = {}
        for line in header_lines:
            split = line.split(",")[0].split("=", maxsplit=1)
            if len(split) != 2:
                continue
            header_dict[split[0].strip()] = split[1].strip()

        self.header = SeriesData(pd.Series(header_dict))
        self.data = assert_not_none(
            reader.pop_csv_block_as_df(header="infer"), "Cannot parse empty dataset"
        )



from allotropy.allotrope.models.adm.pcr.benchling._2023._09.qpcr import Model
from allotropy.allotrope.schema_mappers.adm.pcr.BENCHLING._2023._09.qpcr import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_data_creator import (
    create_calculated_data,
    create_data,
    create_measurement_groups,
    create_metadata,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.vendor_parser import VendorParser


class AppBioQuantStudioDesignandanalysisParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "AppBio QuantStudio Design & Analysis"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = DesignQuantstudioReader.SUPPORTED_EXTENSIONS
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        data = create_data(DesignQuantstudioReader.create(named_file_contents))
        return Data(
            create_metadata(
                data.header,
                named_file_contents.original_file_path,
                data.experiment_type,
            ),
            create_measurement_groups(data.wells.wells, data.header),
            create_calculated_data(data),
        )



from dataclasses import dataclass

from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.revvity_kaleido.kaleido_structure import (
    AnalysisResult,
    BackgroundInfo,
    Data,
    MeasurementInfo,
    Measurements,
    Platemap,
    Results,
)
from allotropy.parsers.utils.values import assert_not_none


@dataclass(frozen=True)
class ResultsV2(Results):
    @classmethod
    def read_barcode(cls, reader: CsvReader) -> str:
        barcode_line = assert_not_none(
            reader.drop_until_inclusive("^Barcode:(.+),"),
            msg="Unable to find background information.",
        )

        raw_barcode, *_ = barcode_line.split(",")
        return raw_barcode.removeprefix("Barcode:").strip()


def create_data_v2(version: str, reader: CsvReader) -> Data:
    return Data(
        version,
        BackgroundInfo.create(reader),
        ResultsV2.create(reader),
        AnalysisResult.create_results(reader, "Measurement Basic Information"),
        MeasurementInfo.create(reader, "Measurement Basic Information", "Plate Type"),
        Platemap.create(reader),
        Measurements.create(reader, "Measurements", "Analysis"),
    )



from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.reader import (
    DataCubeReader,
)
from allotropy.parsers.cytiva_unicorn.structure.data_cube.transformations import (
    Min2Sec,
    Transformation,
)


def create_data_cube(
    handler: UnicornZipHandler,
    label: str,
    data_cube_component: DataCubeComponent,
    transformation: Transformation | None,
) -> DataCube:
    return DataCube(
        label=label,
        structure_dimensions=[
            DataCubeComponent(
                type_=FieldComponentDatatype.float,
                concept="retention time",
                unit="s",
            ),
        ],
        structure_measures=[data_cube_component],
        dimensions=[
            list(
                DataCubeReader(
                    handler=handler,
                    name="Volumes",
                    transformation=Min2Sec(),
                ).get_data()
            )
        ],
        measures=[
            list(
                DataCubeReader(
                    handler=handler,
                    name="Amplitudes",
                    transformation=transformation,
                ).get_data()
            )
        ],
    )



from collections.abc import Iterator
from io import StringIO

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.lines_reader import CsvReader, read_to_lines
from allotropy.parsers.utils.pandas import df_to_series_data, read_csv, SeriesData
from allotropy.parsers.utils.values import assert_not_none


class MethodicalMindReader:
    SUPPORTED_EXTENSIONS = "txt"
    plate_headers: list[SeriesData]
    plate_data: list[pd.DataFrame]

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        file_lines = read_to_lines(named_file_contents)
        reader = CsvReader(file_lines)

        self.plate_headers = []
        self.plate_data = []
        while reader.current_line_exists():
            lines = list(reader.pop_until("Data"))
            lines = [line for line in lines if "Digital_Signature" not in line]
            header_df = read_csv(
                StringIO("\n".join(lines)),
                sep=r":\t",
                engine="python",
                header=None,
                index_col=0,
            ).T
            self.plate_headers.append(df_to_series_data(header_df))

            # Pop data title line
            reader.pop()

            # Create dataframe of the plate, drop any columns with all Nan values
            data = assert_not_none(
                reader.lines_as_df(
                    list(reader.pop_until("===")), sep="\t", header=0, index_col=0
                ),
                "Luminescence data table",
            ).dropna(axis="columns", how="all")
            # There may be multiple rows per well row for additional measurements, and the extra rows are
            # not labelled. ffill the row label so that each row has the corresponding row label.
            data.index = pd.Index(data.index.to_series().ffill())

            data.index = data.index.astype(str).str.strip()
            data.index = pd.Index(data.index.to_series().replace("", pd.NA).ffill())
            data.columns = data.columns.astype(str).str.strip()
            self.plate_data.append(data)

            reader.drop_until("Stack ID")

    def __iter__(self) -> Iterator[tuple[SeriesData, pd.DataFrame]]:
        yield from zip(self.plate_headers, self.plate_data, strict=True)



# generated by datamodel-codegen:
#   filename:  cell-counting.schema.json
#   timestamp: 2024-06-04T03:30:31+00:00

from __future__ import annotations

from dataclasses import dataclass

from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueCell,
    TQuantityValueMicroliter,
    TQuantityValueMicrometer,
    TQuantityValueMillionCellsPerMilliliter,
    TQuantityValueMilliSecond,
    TQuantityValueNanometer,
    TQuantityValuePercent,
    TQuantityValueTODO,
    TQuantityValueUnitless,
)
from allotropy.allotrope.models.shared.definitions.definitions import (
    TDatacubeData,
    TDatacubeStructure,
    TDateTimeValue,
    TStringValue,
)


@dataclass(kw_only=True)
class DeviceSystemDocument:
    device_identifier: TStringValue | None = None
    model_number: TStringValue | None = None
    device_serial_number: TStringValue | None = None


@dataclass(kw_only=True)
class DataSystemDocument:
    data_system_instance_identifier: TStringValue | None = None
    file_name: TStringValue | None = None
    UNC_path: TStringValue | None = None
    software_name: TStringValue | None = None
    software_version: TStringValue | None = None
    ASM_converter_name: TStringValue | None = None
    ASM_converter_version: TStringValue | None = None


@dataclass(kw_only=True)
class SampleDocument:
    sample_identifier: TStringValue | None = None
    batch_identifier: TStringValue | None = None
    sample_role_type: TStringValue | None = None
    location_identifier: TStringValue | None = None


@dataclass(kw_only=True)
class DeviceControlDocumentItem:
    device_type: TStringValue
    sample_volume_setting: TQuantityValueMicroliter | None = None
    detection_type: TStringValue | None = None
    excitation_wavelength_setting: TQuantityValueNanometer | None = None
    emission_wavelength_setting: TQuantityValueNanometer | None = None
    illumination_setting: TQuantityValuePercent | None = None
    exposure_duration_setting: TQuantityValueMilliSecond | None = None
    detector_gain_setting: TStringValue | None = None
    fluorescence_tag_setting: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class DeviceControlAggregateDocument:
    device_control_document: list[DeviceControlDocumentItem]


@dataclass(kw_only=True)
class DataProcessingDocument:
    cell_type_processing_method: TStringValue | None = None
    minimum_cell_diameter: TQuantityValueMicrometer | None = None
    maximum_cell_diameter: TQuantityValueMicrometer | None = None
    cell_density_dilution_factor: TQuantityValueUnitless | None = None


@dataclass(kw_only=True)
class ImageDocumentItem:
    experimental_data_identifier: TStringValue | None = None
    field_index: int | None = None


@dataclass(kw_only=True)
class ImageAggregateDocument:
    image_document: list[ImageDocumentItem] | None = None


@dataclass(kw_only=True)
class TotalCellDiameterDistributionDataCube:
    label: str | None = None
    cube_structure: TDatacubeStructure | None = None
    data: TDatacubeData | None = None


@dataclass(kw_only=True)
class ProcessedDataDocument:
    data_processing_document: DataProcessingDocument
    viability__cell_counter_: TQuantityValuePercent
    viable_cell_density__cell_counter_: TQuantityValueMillionCellsPerMilliliter
    total_cell_count: TQuantityValueCell
    total_cell_density__cell_counter_: TQuantityValueMillionCellsPerMilliliter | None = (
        None
    )
    dead_cell_density__cell_counter_: TQuantityValueMillionCellsPerMilliliter | None = (
        None
    )
    average_total_cell_diameter: TQuantityValueMicrometer | None = None
    average_live_cell_diameter__cell_counter_: TQuantityValueMicrometer | None = None
    average_dead_cell_diameter__cell_counter_: TQuantityValueMicrometer | None = None
    total_cell_diameter_distribution_data_cube: TotalCellDiameterDistributionDataCube | None = (
        None
    )
    viable_cell_count: TQuantityValueCell | None = None
    dead_cell_count: TQuantityValueCell | None = None
    average_total_cell_circularity: TQuantityValueUnitless | None = None
    average_viable_cell_circularity: TQuantityValueUnitless | None = None
    fluorescence_tag_positive_cell_count: TQuantityValueTODO | None = None
    fluorescence_tag_positive_cell_density: TQuantityValueTODO | None = None
    fluorescence_tag_positive_cell_percentage: TQuantityValueTODO | None = None


@dataclass(kw_only=True)
class MeasurementDocumentItem:
    measurement_identifier: TStringValue
    measurement_time: TDateTimeValue
    sample_document: SampleDocument
    device_control_aggregate_document: DeviceControlAggregateDocument
    processed_data_document: ProcessedDataDocument
    image_aggregate_document: ImageAggregateDocument | None = None


@dataclass(kw_only=True)
class MeasurementAggregateDocument:
    measurement_document: list[MeasurementDocumentItem]


@dataclass(kw_only=True)
class CellCountingDocumentItem:
    analyst: TStringValue
    measurement_aggregate_document: MeasurementAggregateDocument
    submitter: TStringValue | None = None


@dataclass(kw_only=True)
class CellCountingAggregateDocument:
    device_system_document: DeviceSystemDocument
    cell_counting_document: list[CellCountingDocumentItem]
    data_system_document: DataSystemDocument | None = None


@dataclass(kw_only=True)
class Model:
    cell_counting_aggregate_document: CellCountingAggregateDocument
    manifest: str = "http://purl.allotrope.org/manifests/cell-counting/BENCHLING/2023/09/cell-counting.manifest"



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path

import pandas as pd

from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
)
from allotropy.exceptions import AllotropyParserError
from allotropy.parsers.constants import DEFAULT_EPOCH_TIMESTAMP, NOT_APPLICABLE
from allotropy.parsers.msd_workbench.constants import (
    DETECTION_TYPE,
    DEVICE_TYPE,
    POSSIBLE_WELL_COUNTS,
    SAMPLE_ROLE_TYPE_MAPPING,
    SOFTWARE_NAME,
)
from allotropy.parsers.utils.pandas import map_rows, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass(frozen=True)
class PlateData:
    measurement_time: str
    plate_well_count: int
    well_plate_id: str
    well_data: pd.DataFrame

    @staticmethod
    def create(
        data: pd.DataFrame,
        well_plate_id: str,
    ) -> PlateData:
        plate_well_count = PlateData._get_plate_well_count(data)
        if not plate_well_count:
            msg = "Could not determine plate well count"
            raise AllotropyParserError(msg)

        return PlateData(
            measurement_time=DEFAULT_EPOCH_TIMESTAMP,
            well_plate_id=well_plate_id,
            plate_well_count=plate_well_count,
            well_data=data,
        )

    @staticmethod
    def _get_plate_well_count(data: pd.DataFrame) -> int | None:
        # Get well numbers via Well ID (1, 2, 3, ...) and well location (A1, B1, ...)
        well_ids = [int(well[-1]) for well in data["Well"]]
        well_location = data["Well"]
        largest_column = sorted([str(loc[0]) for loc in well_location])[-1]
        largest_row = sorted(int(loc[1:]) for loc in well_location)[-1]
        well_number_by_position = (
            ord(largest_column.upper()) - ord("A") + 1
        ) * largest_row
        largest_well_number = max(sorted(well_ids)[-1], well_number_by_position)

        # Round up to the first possible well count GTE the count e.g:
        # - If we have well id 94 but none greater than 96, it's a 96-well plate
        for possible_count in POSSIBLE_WELL_COUNTS:
            if largest_well_number > possible_count:
                continue
            return possible_count
        return None


def create_metadata(file_name: str) -> Metadata:
    asm_file_identifier = Path(file_name).with_suffix(".json")
    return Metadata(
        file_name=file_name.rsplit("\\", 1)[-1],
        unc_path=file_name,
        software_name=SOFTWARE_NAME,
        device_identifier=NOT_APPLICABLE,
        model_number=NOT_APPLICABLE,
        asm_file_identifier=asm_file_identifier.name,
        data_system_instance_id=NOT_APPLICABLE,
    )


def create_measurement_groups(plate_data: PlateData) -> list[MeasurementGroup]:
    def map_measurement(row: SeriesData) -> Measurement:
        sample_id = f"{row[str, 'Sample']}_{row[str, 'Well']}"
        return Measurement(
            type_=MeasurementType.LUMINESCENCE,
            identifier=random_uuid_str(),
            luminescence=row[int, "Signal"],
            sample_identifier=sample_id,
            location_identifier=row[str, "Spot"],
            well_location_identifier=row[str, "Well"],
            well_plate_identifier=plate_data.well_plate_id,
            device_type=DEVICE_TYPE,
            detection_type=DETECTION_TYPE,
            mass_concentration=row.get(float, "Concentration"),
            sample_role_type=SAMPLE_ROLE_TYPE_MAPPING.get(
                row[str, "Sample"][0].lower()
            ),
            measurement_custom_info={
                "detection range": row.get(str, "Detection Range"),
                "assay identifier": row.get(str, "Assay"),
            },
            sample_custom_info={
                "dilution factor setting": row.get(int, "Dilution Factor"),
            },
        )

    measurements = map_rows(plate_data.well_data, map_measurement)

    grouped_measurements = defaultdict(list)
    for measurement in measurements:
        grouped_measurements[measurement.sample_identifier].append(measurement)

    return [
        MeasurementGroup(
            measurement_time=plate_data.measurement_time,
            plate_well_count=plate_data.plate_well_count,
            measurements=group,
        )
        for group in grouped_measurements.values()
    ]



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from io import StringIO
import math
from pathlib import Path
import re

import pandas as pd

from allotropy.allotrope.models.shared.definitions.definitions import (
    FieldComponentDatatype,
)
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    CalculatedDataItem,
    DataSource,
    ErrorDocument,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ScanPositionSettingPlateReader,
)
from allotropy.allotrope.schema_mappers.data_cube import DataCube, DataCubeComponent
from allotropy.exceptions import (
    AllotropeConversionError,
    AllotropyParserError,
)
from allotropy.parsers.agilent_gen5.agilent_gen5_reader import AgilentGen5Reader
from allotropy.parsers.agilent_gen5.constants import (
    ALPHALISA_FLUORESCENCE_FOUND,
    DATA_SOURCE_FEATURE_VALUES,
    DEFAULT_SOFTWARE_NAME,
    DEVICE_TYPE,
    ELAPSED_TIME,
    EMISSION_KEY,
    EXCITATION_KEY,
    FILENAME_REGEX,
    GAIN_KEY,
    MEASUREMENTS_DATA_POINT_KEY,
    MIRROR_KEY,
    NAN_EMISSION_EXCITATION,
    OPTICS_KEY,
    PATHLENGTH_CORRECTION_KEY,
    READ_DATA_MEASUREMENT_ERROR,
    READ_HEIGHT_KEY,
    READ_MODE_UNITS,
    READ_SPEED_KEY,
    ReadMode,
    ReadType,
    SECONDS,
    UNSUPPORTED_READ_MODE_ERROR,
    UNSUPPORTED_READ_TYPE_ERROR,
    WAVELENGTHS_KEY,
)
from allotropy.parsers.agilent_gen5_image.constants import POSSIBLE_WELL_COUNTS
from allotropy.parsers.constants import NEGATIVE_ZERO, NOT_APPLICABLE
from allotropy.parsers.lines_reader import SectionLinesReader
from allotropy.parsers.utils.pandas import read_csv, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    try_float,
    try_float_or_none,
    try_non_nan_float_or_none,
)


@dataclass(frozen=True)
class HeaderData:
    datetime: str
    software_version: str
    experiment_file_path: str | None
    protocol_file_path: str | None
    well_plate_identifier: str | None
    model_number: str | None
    equipment_serial_number: str | None
    plate_well_count: float
    file_name: str
    unc_path: str

    @classmethod
    def create(cls, data: SeriesData, file_path: str) -> HeaderData:
        file_name = Path(file_path).name
        matches = re.match(FILENAME_REGEX, file_name)
        plate_identifier = matches.groupdict()["plate_identifier"] if matches else None
        plate_well_count = cls._extract_well_count(data[str, "Plate Type"])
        return HeaderData(
            software_version=data[str, "Software Version"],
            experiment_file_path=data.get(str, "Experiment File Path:"),
            file_name=file_name,
            unc_path=file_path,
            protocol_file_path=data.get(str, "Protocol File Path:"),
            datetime=f'{data[str, "Date"]} {data[str, "Time"]}',
            well_plate_identifier=plate_identifier or data.get(str, "Plate Number"),
            model_number=data.get(str, "Reader Type:"),
            equipment_serial_number=data.get(str, "Reader Serial Number:"),
            plate_well_count=plate_well_count,
        )

    @staticmethod
    def _extract_well_count(plate_type: str) -> float:
        match = re.search(
            rf"({'|'.join(str(count) for count in POSSIBLE_WELL_COUNTS)})", plate_type
        )
        if match:
            return float(match.group())
        return 0


@dataclass(frozen=True)
class FilterSet:
    gain: str
    emission: str | None = None
    excitation: str | None = None
    mirror: str | None = None
    optics: str | None = None

    @property
    def detector_wavelength_setting(self) -> float | None:
        if not self.emission:
            return None
        return try_non_nan_float_or_none(self.emission.split("/")[0])

    @property
    def detector_bandwidth_setting(self) -> float | None:
        if not self.emission:
            return None
        try:
            return try_non_nan_float_or_none(self.emission.split("/")[1])
        except IndexError:
            return None

    @property
    def excitation_wavelength_setting(self) -> float | None:
        if self.excitation:
            return try_non_nan_float_or_none(self.excitation.split("/")[0])
        return None

    @property
    def excitation_bandwidth_setting(self) -> float | None:
        if not self.excitation:
            return None
        try:
            return try_float(self.excitation.split("/")[1], "Excitation bandwith")
        except IndexError:
            return None

    @property
    def wavelength_filter_cutoff_setting(self) -> float | None:
        if self.mirror:
            return try_float(self.mirror.split(" ")[1], "Wavelength filter cutoff")
        return None

    @property
    def scan_position_setting(self) -> ScanPositionSettingPlateReader | None:
        position_lookup = {
            "Top": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
            "Bottom": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
        }
        position = self.optics
        if self.mirror:
            position = self.mirror.split(" ")[0]

        # TODO: Raise if position is not a valid lookup value
        return position_lookup.get(position) if position is not None else None


@dataclass
class DeviceControlData:
    step_label: str | None
    _list_data: dict[str, list[str]]
    _single_data: dict[str, str]

    LIST_KEYS = frozenset(
        {
            EMISSION_KEY,
            EXCITATION_KEY,
            OPTICS_KEY,
            GAIN_KEY,
            MIRROR_KEY,
            WAVELENGTHS_KEY,
        }
    )

    def __init__(self) -> None:
        self.step_label = None
        self._list_data = {key: [] for key in self.LIST_KEYS}
        self._single_data = {}

    def add(self, key: str, value: str | list[str]) -> None:
        if key in self.LIST_KEYS:
            if isinstance(value, str):
                self._list_data[key].append(value)
            else:
                self._list_data[key].extend(value)
        else:
            if not isinstance(value, str):
                msg = f"Unexpected list passed to key: {key}"
                raise AllotropyParserError(msg)
            self._single_data[key] = value

    def get_list(self, key: str) -> list[str]:
        return self._list_data.get(key, [])

    def get(self, key: str) -> str | None:
        return self._single_data.get(key)

    @classmethod
    def create(cls, lines: list[str], read_mode: ReadMode) -> DeviceControlData:

        device_control_data = DeviceControlData()

        for line in lines:
            strp_line = str(line.strip())
            if strp_line.startswith("Read\t"):
                device_control_data.step_label = cls._get_step_label(line, read_mode)
                continue

            elif strp_line.startswith(WAVELENGTHS_KEY):
                wavelengths = strp_line.split(":  ")
                device_control_data.add(WAVELENGTHS_KEY, wavelengths[1].split(", "))
                continue

            line_data: list[str] = strp_line.split(",  ")
            for read_datum in line_data:
                splitted_datum = read_datum.split(": ")
                if len(splitted_datum) != 2:
                    continue
                device_control_data.add(splitted_datum[0], splitted_datum[1])

        return device_control_data

    @classmethod
    def _get_step_label(cls, read_line: str, read_mode: str) -> str | None:
        split_line = read_line.strip().split("\t")
        if len(split_line) != 2:
            msg = (
                f"Expected the Read data line {split_line} to contain exactly 2 values."
            )
            raise AllotropeConversionError(msg)
        if split_line[1] != f"{read_mode.title()} Endpoint":
            return split_line[1]

        return None


@dataclass(frozen=True)
class KineticData:
    run_time: str | None
    interval: str | None
    reads: int | None

    @classmethod
    def create(cls, lines: list[str]) -> KineticData | None:
        for line in lines:
            if line.startswith("Start Kinetic\t"):
                return cls._parse_kinetic_section(line)
        return None

    @classmethod
    def _parse_kinetic_section(cls, line: str) -> KineticData:
        split_line = line.strip().split("\t")[1].split(", ")
        run_value = [part for part in split_line if "Runtime" in part]
        if len(run_value) != 1:
            msg = f"Could not find 'Runtime' in Kinetic line: {line}."
            raise AllotropeConversionError(msg)
        interval = [part for part in split_line if "Interval" in part]
        if len(interval) != 1:
            msg = f"Could not find 'Interval' in Kinetic line: {line}."
            raise AllotropeConversionError(msg)
        reads = [part for part in split_line if "Reads" in part]
        if len(reads) != 1:
            msg = f"Could not find 'Reads' in Kinetic line: {line}."
            raise AllotropeConversionError(msg)
        return KineticData(
            run_time=run_value[0].split(" ")[1].strip(),
            interval=interval[0].split(" ")[1].strip(),
            reads=int(reads[0].split(" ")[0].strip()),
        )


@dataclass(frozen=True)
class ReadData:
    read_mode: ReadMode
    measurement_labels: set[str]
    pathlength_correction: str | None
    step_label: str | None
    number_of_averages: float | None
    detector_distance: float | None
    detector_carriage_speed: str | None
    filter_sets: dict[str, FilterSet]

    @classmethod
    def create(cls, lines: list[str]) -> list[ReadData]:
        procedure_details = "\n".join(lines)
        read_type = cls.get_read_type(procedure_details)
        if read_type != ReadType.ENDPOINT:
            raise AllotropeConversionError(UNSUPPORTED_READ_TYPE_ERROR)
        read_modes = cls.get_read_modes(procedure_details)
        read_sections = list(SectionLinesReader(lines).iter_sections(r"^\s*Read\t"))
        if len(read_modes) != len(read_sections):
            msg = "Expected the number of read modes to match the number of read sections."
            raise AllotropeConversionError(msg)

        read_data_list: list[ReadData] = []
        for read_mode, read_section in zip(read_modes, read_sections, strict=True):
            device_control_data = DeviceControlData.create(
                read_section.lines, read_mode
            )
            measurement_labels, label_aliases = cls._get_measurement_labels(
                device_control_data, read_mode
            )
            all_labels = {
                alias for aliases in label_aliases.values() for alias in aliases
            }
            number_of_averages = device_control_data.get(MEASUREMENTS_DATA_POINT_KEY)
            read_height = device_control_data.get(READ_HEIGHT_KEY) or ""
            read_data_list.append(
                ReadData(
                    read_mode=read_mode,
                    step_label=device_control_data.step_label,
                    measurement_labels=set(measurement_labels) | all_labels,
                    detector_carriage_speed=device_control_data.get(READ_SPEED_KEY),
                    # Absorbance attributes
                    pathlength_correction=device_control_data.get(
                        PATHLENGTH_CORRECTION_KEY
                    ),
                    number_of_averages=try_float_or_none(number_of_averages),
                    # Luminescence attributes
                    detector_distance=try_float_or_none(read_height.split(" ")[0]),
                    # Fluorescence attributes
                    filter_sets=cls._get_filter_sets(
                        measurement_labels,
                        label_aliases,
                        device_control_data,
                        read_mode,
                    ),
                )
            )
        return read_data_list

    @staticmethod
    def get_read_modes(procedure_details: str) -> list[ReadMode]:
        read_modes = []
        for read_mode in ReadMode:
            # Construct the regex pattern for the current read mode
            pattern = fr"\t{re.escape(read_mode.value)} Endpoint"
            # Use regex to find all occurrences of the read mode pattern in the procedure details
            matches = re.findall(pattern, procedure_details)
            if matches:
                # Add the read_mode to the list for each match found
                read_modes.extend([read_mode] * len(matches))

        if not read_modes:
            raise AllotropeConversionError(UNSUPPORTED_READ_MODE_ERROR)

        if ReadMode.ALPHALISA in read_modes and ReadMode.FLUORESCENCE in read_modes:
            raise AllotropeConversionError(ALPHALISA_FLUORESCENCE_FOUND)

        # Replace ALPHALISA with FLUORESCENCE
        read_modes = [
            ReadMode.FLUORESCENCE if read_mode == ReadMode.ALPHALISA else read_mode
            for read_mode in read_modes
        ]

        return read_modes

    @staticmethod
    def get_read_type(procedure_details: str) -> ReadType:
        if ReadType.AREASCAN.value in procedure_details:
            return ReadType.AREASCAN
        elif ReadType.SPECTRAL.value in procedure_details:
            return ReadType.SPECTRAL
        # check for this last, because other modes still contain the word "Endpoint"
        elif ReadType.ENDPOINT.value in procedure_details:
            return ReadType.ENDPOINT

        msg = f"Read type not found; expected to find one of {sorted(ReadType._member_names_)}."
        raise AllotropeConversionError(msg)

    @classmethod
    def _get_measurement_labels(
        cls, device_control_data: DeviceControlData, read_mode: str
    ) -> tuple[list[str], dict[str, set[str]]]:
        step_label = device_control_data.step_label
        label_prefix = f"{step_label}:" if step_label else ""
        measurement_labels = []
        # Some measurement labels may be reported in more than one format in the result rows, e.g.
        # fluorescence measurements may include bandwidths, or not: 360/40,460/40 or 360,460.
        label_aliases: dict[str, set[str]] = {}
        if read_mode == ReadMode.ABSORBANCE:
            measurement_labels = cls._get_absorbance_measurement_labels(
                label_prefix, device_control_data
            )

        if read_mode == ReadMode.FLUORESCENCE:
            excitations: list[str] = device_control_data.get_list(EXCITATION_KEY)
            emissions: list[str] = device_control_data.get_list(EMISSION_KEY)
            measurement_labels = [
                f"{label_prefix}{excitation},{emission}"
                for excitation, emission in zip(excitations, emissions, strict=True)
            ]
            label_aliases = {
                f"{label_prefix}{excitation},{emission}": {
                    f"{label_prefix}{excitation.split('/')[0]},{emission.split('/')[0]}"
                }
                for excitation, emission in zip(excitations, emissions, strict=True)
            }
            if not measurement_labels:
                measurement_labels = ["Alpha"]

        if read_mode == ReadMode.LUMINESCENCE:
            emissions = device_control_data.get_list(EMISSION_KEY)
            for emission in emissions:
                label = "Lum" if emission in NAN_EMISSION_EXCITATION else emission
                measurement_labels.append(f"{label_prefix}{label}")

        return measurement_labels, label_aliases

    @classmethod
    def _get_absorbance_measurement_labels(
        cls, label_prefix: str | None, device_control_data: DeviceControlData
    ) -> list[str]:
        pathlength_correction = device_control_data.get(PATHLENGTH_CORRECTION_KEY)
        measurement_labels = []

        for wavelength in device_control_data.get_list(WAVELENGTHS_KEY):
            label = f"{label_prefix}{wavelength}"
            measurement_labels.append(label)

        if pathlength_correction:
            test, ref = pathlength_correction.split(" / ")
            test_label = f"{label_prefix}{test} [Test]"
            ref_label = f"{label_prefix}{ref} [Ref]"
            measurement_labels.extend([test_label, ref_label])

        return measurement_labels

    @classmethod
    def _get_filter_sets(
        cls,
        measurement_labels: list[str],
        label_aliases: dict[str, set[str]],
        device_control_data: DeviceControlData,
        read_mode: ReadMode,
    ) -> dict[str, FilterSet]:
        filter_data: dict[str, FilterSet] = {}
        if read_mode == ReadMode.ABSORBANCE:
            return filter_data

        emissions = device_control_data.get_list(EMISSION_KEY)
        excitations = device_control_data.get_list(EXCITATION_KEY)
        mirrors = device_control_data.get_list(MIRROR_KEY)
        optics = device_control_data.get_list(OPTICS_KEY)
        gains = device_control_data.get_list(GAIN_KEY)

        if len(measurement_labels) != len(gains):
            msg = f"Expected the number measurement labels: {measurement_labels} to match the number of gains: {gains}."
            raise AllotropeConversionError(msg)

        for idx, label in enumerate(measurement_labels):
            mirror = None
            if mirrors and read_mode == ReadMode.FLUORESCENCE:
                mirror = mirrors[idx]
            filter_data[label] = FilterSet(
                emission=emissions[idx] if emissions else None,
                gain=gains[idx],
                excitation=excitations[idx] if excitations else None,
                mirror=mirror,
                optics=optics[idx] if optics else None,
            )
        for measurement_label, aliases in label_aliases.items():
            for alias in aliases:
                filter_data[alias] = filter_data[measurement_label]

        return filter_data


def _validate_result_sections(result_sections: list[list[str]]) -> None:
    """Validates whether all the result sections dimensions are consistent."""
    first_section = result_sections[0]

    for section in result_sections[1:]:
        if not first_section[0] == section[0] and len(first_section) == len(section):
            msg = "All result tables should have the same dimensions."
            raise AllotropeConversionError(msg)


def get_results_section(reader: AgilentGen5Reader) -> list[str] | None:
    """Returns a valid Results Matrix from the reader sections if found.

    Checks for Results in the reader sections, if not found, creates the results matrix with all
    sections that are correctly formatted as a results table (excluding the Layout section). If
    no tables with results are found, returns None
    """
    if "Results" in reader.sections:
        return reader.sections["Results"]

    def is_results(section: list[str]) -> bool:
        return (
            len(section) > 2
            and section[1].startswith("\t1")
            and section[2].startswith("A\t")
        )

    result_sections = []
    for name, section in reader.sections.items():
        if name == "Layout":
            continue
        if is_results(section):
            result_sections.append(section[1:])

    if result_sections:
        _validate_result_sections(result_sections)
        return [
            "Results",
            result_sections[0][0],
            *[
                section[i + 1]
                for i in range(len(result_sections[0]) - 1)
                for section in result_sections
            ],
        ]

    return None


def get_identifiers(layout_lines: list[str] | None) -> dict[str, str]:
    if not layout_lines:
        return {}
    # Create dataframe from tabular data and forward fill empty values in index
    data = read_csv(StringIO("\n".join(layout_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis="index").values)

    identifiers = {}
    for row_name, row in data.iterrows():
        label = row.iloc[-1]
        for col_index, col in enumerate(row.iloc[:-1]):
            well_pos = f"{row_name}{col_index + 1}"
            # Prefer Name to Well ID
            if not pd.isna(col) and (
                label == "Name" or label == "Well ID" and well_pos not in identifiers
            ):
                identifiers[well_pos] = col
    return identifiers


def get_temperature(actual_temperature_lines: list[str] | None) -> float | None:
    if not actual_temperature_lines:
        return None
    if len(actual_temperature_lines) != 1:
        msg = f"Expected the Temperature section '{actual_temperature_lines}' to contain exactly 1 line."
        raise AllotropeConversionError(msg)

    return try_float(
        actual_temperature_lines[0].strip().split("\t")[-1], "Actual Temperature"
    )


def get_kinetic_measurements(
    kinetic_lines: list[str] | None,
) -> tuple[dict[str, list[float | None]], list[float]] | None:
    if not kinetic_lines:
        return None
    data = (
        read_csv(StringIO("\n".join(kinetic_lines)), sep="\t", index_col=0)
        .dropna(axis="columns", how="all")
        .dropna(axis="index", how="all")
    )

    kinetic_measurements: defaultdict[str, list[float | None]] = defaultdict(
        list[float | None]
    )
    kinetic_elapsed_time: list[float] = data.index.map(
        lambda val: _convert_time_to_seconds(str(val))
    ).to_list()
    for col_name, column in data.items():
        if col_name == "Time":
            continue
        if column.isnull().any():
            msg = f"Unable to process null value in column: {col_name} of Kinetic section."
            raise AllotropeConversionError(msg)
        kinetic_measurements[str(col_name)] = [
            float(value) if not pd.isna(value) else None for value in column
        ]
    return dict(kinetic_measurements.items()), kinetic_elapsed_time


@dataclass(frozen=True)
class MeasurementData:
    identifier: str
    value: float | None
    label: str


def create_results(
    result_lines: list[str],
    header_data: HeaderData,
    read_data: list[ReadData],
    sample_identifiers: dict[str, str],
    actual_temperature: float | None,
) -> tuple[list[MeasurementGroup], list[CalculatedDataItem]]:
    if result_lines[0].strip() != "Results":
        msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
        raise AllotropeConversionError(msg)

    # Create dataframe from tabular data and forward fill empty values in index
    data = read_csv(StringIO("\n".join(result_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis="index").values)

    well_to_measurements: defaultdict[str, list[MeasurementData]] = defaultdict(
        list[MeasurementData]
    )
    calculated_data: defaultdict[str, list[tuple[str, float]]] = defaultdict(
        list[tuple[str, float]]
    )
    measurement_labels = [
        label for r_data in read_data for label in r_data.measurement_labels
    ]
    error_documents_per_well: defaultdict[str, list[ErrorDocument]] = defaultdict(
        list[ErrorDocument]
    )
    for row_name, row in data.iterrows():
        label = str(row.iloc[-1])
        for col_index, value in enumerate(row.iloc[:-1]):
            well_pos = f"{row_name}{col_index + 1}"
            well_value = try_float_or_none(value)
            is_measurement = label in measurement_labels
            # skip empty values
            if isinstance(well_value, float) and math.isnan(well_value):
                continue

            # Report error documents for NaN values
            if well_value is None:
                error_documents_per_well[well_pos].append(
                    ErrorDocument(
                        error=value,
                        # TODO Add support for multiple read modes
                        error_feature=(
                            read_data[0].read_mode.lower() if is_measurement else label
                        ),
                    )
                )
            if is_measurement:

                well_to_measurements[well_pos].append(
                    MeasurementData(
                        random_uuid_str(),
                        NEGATIVE_ZERO if well_value is None else well_value,
                        label,
                    )
                )
            elif well_value is not None:
                calculated_data[well_pos].append((label, well_value))

    groups = [
        MeasurementGroup(
            measurement_time=header_data.datetime,
            plate_well_count=len(set(data.index.tolist()))
            * len(set(data.columns[1:].tolist())),
            analytical_method_identifier=header_data.protocol_file_path,
            experimental_data_identifier=header_data.experiment_file_path,
            measurements=[
                _create_measurement(
                    measurement,
                    well_position,
                    header_data,
                    get_read_data_from_measurement(measurement, read_data),
                    sample_identifiers.get(well_position),
                    actual_temperature,
                    error_documents=error_documents_per_well.get(well_position),
                )
                for measurement in measurements
            ],
        )
        for well_position, measurements in well_to_measurements.items()
    ]

    calculated_data_items = [
        CalculatedDataItem(
            identifier=random_uuid_str(),
            data_sources=[
                DataSource(
                    identifier=measurement.identifier,
                    feature=item.read_mode.value.lower(),
                )
                for measurement in _get_sources(
                    label, well_to_measurements[well_position]
                )
                for item in read_data
            ],
            unit=UNITLESS,
            name=label,
            value=value,
        )
        for well_position, well_calculated_data in calculated_data.items()
        for label, value in well_calculated_data
    ]

    return groups, calculated_data_items


def create_kinetic_results(
    result_lines: list[str],
    header_data: HeaderData,
    read_data: list[ReadData],
    sample_identifiers: dict[str, str],
    actual_temperature: float | None,
    kinetic_data: KineticData,
    kinetic_measurements: dict[str, list[float | None]],
    kinetic_elapsed_time: list[float],
) -> tuple[list[MeasurementGroup], list[CalculatedDataItem]]:
    if result_lines[0].strip() != "Results":
        msg = f"Expected the first line of the results section '{result_lines[0]}' to be 'Results'."
        raise AllotropeConversionError(msg)

    # Create dataframe from tabular data and forward fill empty values in index
    data = read_csv(StringIO("\n".join(result_lines[1:])), sep="\t")
    data = data.set_index(data.index.to_series().ffill(axis="index").values)

    calculated_data: defaultdict[str, list[tuple[str, float]]] = defaultdict(
        list[tuple[str, float]]
    )
    measurement_labels = [
        label for r_data in read_data for label in r_data.measurement_labels
    ]
    for row_name, row in data.iterrows():
        label = row.iloc[-1]
        for col_index, value in enumerate(row.iloc[:-1]):
            well_pos = f"{row_name}{col_index + 1}"
            well_value = try_non_nan_float_or_none(value)
            # TODO: Report error documents for NaN values
            if not well_value:
                continue
            calculated_data[well_pos].append((label, well_value))

    groups = [
        MeasurementGroup(
            measurement_time=header_data.datetime,
            plate_well_count=len(set(data.index.tolist()))
            * len(set(data.columns[1:].tolist())),
            analytical_method_identifier=header_data.protocol_file_path,
            experimental_data_identifier=header_data.experiment_file_path,
            measurements=[
                _create_measurement(
                    measurement := MeasurementData(
                        random_uuid_str(), None, measurement_labels[0]
                    ),
                    well_position,
                    header_data,
                    get_read_data_from_measurement(measurement, read_data),
                    sample_identifiers.get(well_position),
                    actual_temperature,
                    kinetic_data,
                    kinetic_measurements[well_position],
                    kinetic_elapsed_time,
                )
            ],
        )
        for well_position in calculated_data.keys()
    ]

    groups_by_well_position = {
        group.measurements[0].location_identifier: group for group in groups
    }

    calculated_data_items = [
        CalculatedDataItem(
            identifier=random_uuid_str(),
            data_sources=[
                DataSource(
                    identifier=groups_by_well_position[well_position]
                    .measurements[0]
                    .identifier,
                    # TODO: Add support for multiple kinetic sections
                    feature=DATA_SOURCE_FEATURE_VALUES[read_data[0].read_mode],
                )
            ],
            unit=UNITLESS,
            name=label,
            value=value,
        )
        for well_position, well_calculated_data in calculated_data.items()
        for label, value in well_calculated_data
    ]

    return groups, calculated_data_items


def get_read_data_from_measurement(
    measurement: MeasurementData, read_data_list: list[ReadData]
) -> ReadData:
    for read_data in read_data_list:
        if _is_label_in_measurement_labels(
            measurement.label, read_data.measurement_labels
        ):
            return read_data

    raise AllotropeConversionError(
        READ_DATA_MEASUREMENT_ERROR.format(measurement.label)
    )


def _get_sources(
    calculated_data_label: str, measurements: list[MeasurementData]
) -> list[MeasurementData]:
    # Pathlength is a special case, its sources are always determined
    # by the pathlength correction setting
    if calculated_data_label.split(":")[-1] == "Pathlength":
        pathlength_suffixes = ["[Test]", "[Ref]"]
        sources = [
            measurement
            for measurement in measurements
            if measurement.label.split(" ")[-1] in pathlength_suffixes
        ]
    else:
        sources = [
            measurement
            for measurement in measurements
            if measurement.label in calculated_data_label
        ]
    # if there are no matches in the measurement labels, use all measurements as source
    return sources or measurements


def create_metadata(header_data: HeaderData) -> Metadata:
    asm_file_identifier = Path(header_data.file_name).with_suffix(".json")
    return Metadata(
        device_identifier=NOT_APPLICABLE,
        model_number=header_data.model_number or NOT_APPLICABLE,
        equipment_serial_number=header_data.equipment_serial_number,
        software_name=DEFAULT_SOFTWARE_NAME,
        software_version=header_data.software_version,
        file_name=header_data.file_name,
        asm_file_identifier=asm_file_identifier.name,
        data_system_instance_id=NOT_APPLICABLE,
        unc_path=header_data.unc_path,
    )


def _is_label_in_measurement_labels(label: str, measurement_labels: set[str]) -> bool:
    if not measurement_labels:
        return False
    if label in measurement_labels:
        return True
    # TODO Improve this logic to support multiple kinetic sections
    match = re.search(r"\[(\d+)]", label)
    if match and match.group()[1:-1] in measurement_labels:
        return True
    return False


def _create_measurement(
    measurement: MeasurementData,
    well_position: str,
    header_data: HeaderData,
    read_data: ReadData,
    sample_identifier: str | None,
    actual_temperature: float | None,
    kinetic_data: KineticData | None = None,
    kinetic_measurements: list[float | None] | None = None,
    kinetic_elapsed_time: list[float] | None = None,
    error_documents: list[ErrorDocument] | None = None,
) -> Measurement:
    if read_data.read_mode == ReadMode.ABSORBANCE and not kinetic_data:
        measurement_type = MeasurementType.ULTRAVIOLET_ABSORBANCE
    elif read_data.read_mode == ReadMode.FLUORESCENCE and not kinetic_data:
        measurement_type = MeasurementType.FLUORESCENCE
    elif read_data.read_mode == ReadMode.LUMINESCENCE and not kinetic_data:
        measurement_type = MeasurementType.LUMINESCENCE
    elif read_data.read_mode == ReadMode.ABSORBANCE and kinetic_data:
        measurement_type = MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR
    elif read_data.read_mode == ReadMode.FLUORESCENCE and kinetic_data:
        measurement_type = MeasurementType.FLUORESCENCE_CUBE_DETECTOR
    elif read_data.read_mode == ReadMode.LUMINESCENCE and kinetic_data:
        measurement_type = MeasurementType.LUMINESCENCE_CUBE_DETECTOR

    if measurement_type in [
        MeasurementType.ULTRAVIOLET_ABSORBANCE,
        MeasurementType.ULTRAVIOLET_ABSORBANCE_CUBE_DETECTOR,
    ]:
        filter_data = None
        detector_wavelength_setting = (
            float(measurement.label.split(":")[-1].split(" ")[0])
        ) or None
    else:
        filter_data = read_data.filter_sets[measurement.label]
        detector_wavelength_setting = filter_data.detector_wavelength_setting

    return Measurement(
        type_=measurement_type,
        device_type=DEVICE_TYPE,
        identifier=measurement.identifier,
        sample_identifier=sample_identifier
        or f"{header_data.well_plate_identifier} {well_position}",
        location_identifier=well_position,
        well_plate_identifier=header_data.well_plate_identifier,
        detection_type=read_data.read_mode.value,
        detector_wavelength_setting=detector_wavelength_setting,
        detector_bandwidth_setting=(
            filter_data.detector_bandwidth_setting if filter_data else None
        ),
        excitation_wavelength_setting=(
            filter_data.excitation_wavelength_setting if filter_data else None
        ),
        excitation_bandwidth_setting=(
            filter_data.excitation_bandwidth_setting if filter_data else None
        ),
        wavelength_filter_cutoff_setting=(
            filter_data.wavelength_filter_cutoff_setting if filter_data else None
        ),
        detector_distance_setting=read_data.detector_distance,
        scan_position_setting=(
            filter_data.scan_position_setting if filter_data else None
        ),
        detector_gain_setting=filter_data.gain if filter_data else None,
        number_of_averages=read_data.number_of_averages,
        detector_carriage_speed=read_data.detector_carriage_speed,
        absorbance=(
            measurement.value
            if measurement_type == MeasurementType.ULTRAVIOLET_ABSORBANCE
            and not kinetic_data
            else None
        ),
        fluorescence=(
            measurement.value
            if measurement_type == MeasurementType.FLUORESCENCE and not kinetic_data
            else None
        ),
        luminescence=(
            measurement.value
            if measurement_type == MeasurementType.LUMINESCENCE and not kinetic_data
            else None
        ),
        compartment_temperature=actual_temperature,
        total_measurement_time_setting=(
            _convert_time_to_seconds(kinetic_data.run_time) if kinetic_data else None
        ),
        read_interval_setting=(
            _convert_time_to_seconds(kinetic_data.interval) if kinetic_data else None
        ),
        number_of_scans_setting=kinetic_data.reads if kinetic_data else None,
        profile_data_cube=(
            DataCube(
                label=str(detector_wavelength_setting),
                structure_dimensions=[
                    DataCubeComponent(
                        concept=ELAPSED_TIME,
                        type_=FieldComponentDatatype.double,
                        unit=SECONDS,
                    )
                ],
                structure_measures=[
                    DataCubeComponent(
                        concept=read_data.read_mode.lower(),
                        type_=FieldComponentDatatype.double,
                        unit=READ_MODE_UNITS.get(read_data.read_mode, UNITLESS),
                    )
                ],
                dimensions=[kinetic_elapsed_time],
                measures=[kinetic_measurements],
            )
            if kinetic_elapsed_time and kinetic_measurements
            else None
        ),
        error_document=error_documents,
    )


def _convert_time_to_seconds(time_str: str | None) -> float:
    if not time_str:
        return 0.0
    try:
        hours, minutes, seconds = map(float, time_str.split(":"))
        return hours * 3600 + minutes * 60 + seconds

    except ValueError:
        msg = f"Invalid time string: '{time_str}'."
        raise AllotropeConversionError(msg) from None



from dataclasses import dataclass

from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    WellList,
)


@dataclass(frozen=True)
class StandardCurveWellList(WellList):
    @classmethod
    def get_data_sheet(cls) -> str:
        return "Standard Curve Result"



from enum import Enum
import re

from allotropy.exceptions import (
    AllotropeConversionError,
    msg_for_error_on_unrecognized_value,
)
from allotropy.parsers.lines_reader import CsvReader, InvertedLinesReader
from allotropy.parsers.revvity_kaleido.kaleido_structure import Data
from allotropy.parsers.revvity_kaleido.kaleido_structure_v2 import create_data_v2
from allotropy.parsers.revvity_kaleido.kaleido_structure_v3 import create_data_v3
from allotropy.parsers.utils.values import assert_not_none


class Version(Enum):
    V2 = "2.0"
    V3 = "3.0"
    V3_5 = "3.5"


def get_version(reader: CsvReader) -> str:
    inv_reader = InvertedLinesReader(reader.lines)
    last_line = assert_not_none(
        inv_reader.pop_data(),
        msg="Unable to find last line of input file.",
    )
    return assert_not_none(
        re.search(r"Exported with Kaleido ([\d.]+)", last_line),
        msg="Unable to find Revvity Kaleido version in input file.",
    ).group(1)


def create_data(reader: CsvReader) -> Data:
    version = get_version(reader)

    if version.startswith(Version.V2.value):
        return create_data_v2(version, reader)
    elif version.startswith(Version.V3.value) or version.startswith(Version.V3_5.value):
        return create_data_v3(version, reader)
    else:
        valid_versions = [f"v{v.value}+" for v in Version]
        msg = msg_for_error_on_unrecognized_value(
            "Revvity Kaleido version", version, valid_versions
        )
        raise AllotropeConversionError(msg)



import pandas as pd

from allotropy.constants import DEFAULT_ENCODING
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    read_csv,
    read_excel,
    SeriesData,
)


class RocheCedexHiResReader:
    SUPPORTED_EXTENSIONS = "csv,xlsx"
    header: SeriesData
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents) -> None:
        if named_file_contents.extension == "csv":
            df = read_csv(
                named_file_contents.contents,
                index_col=False,
                encoding=DEFAULT_ENCODING,
            )
        else:
            df = read_excel(named_file_contents.contents)

        # Fix typo found in some source files.
        df.columns = df.columns.str.replace("identifer", "identifier", regex=True)
        self.data = df
        self.header = df_to_series_data(df.head(1))



from abc import ABC, abstractmethod
from re import search
from typing import ClassVar

from allotropy.parsers.appbio_quantstudio_designandanalysis.appbio_quantstudio_designandanalysis_reader import (
    DesignQuantstudioReader,
)
from allotropy.parsers.appbio_quantstudio_designandanalysis.structure.generic.structure import (
    Data,
)


class Creator(ABC):
    PLUGIN_REGEX: ClassVar[str]
    EXPECTED_SHEETS: ClassVar[list[str]]

    @classmethod
    def check_sheets(cls, reader: DesignQuantstudioReader) -> bool:
        return all(reader.has_sheet(sheet_name) for sheet_name in cls.EXPECTED_SHEETS)

    @classmethod
    def check_plugin_name(cls, raw_plugin_name: str | None) -> bool:
        return (
            True
            if raw_plugin_name is None
            else bool(search(cls.PLUGIN_REGEX, raw_plugin_name))
        )

    @classmethod
    def check_experiment_type(
        cls, reader: DesignQuantstudioReader, raw_plugin_name: str | None
    ) -> bool:
        return cls.check_sheets(reader) and cls.check_plugin_name(raw_plugin_name)

    @classmethod
    @abstractmethod
    def create(cls, reader: DesignQuantstudioReader) -> Data:
        pass



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
import logging
from pathlib import Path
import re

import pandas as pd

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ScanPositionSettingPlateReader,
    TransmittedLightSetting,
)
from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.schema_mappers.adm.plate_reader.benchling._2023._09.plate_reader import (
    DataSource,
    ImageFeature,
    ImageSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ProcessedData,
)
from allotropy.exceptions import AllotropeConversionError, AllotropeParsingError
from allotropy.parsers.lines_reader import CsvReader, EMPTY_STR_OR_CSV_LINE
from allotropy.parsers.revvity_kaleido import constants
from allotropy.parsers.utils.pandas import df_to_series_data, SeriesData
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    try_float,
    try_float_or_none,
)


class ExperimentType(Enum):
    FLUORESCENCE = "fluorescence"
    ABSORBANCE = "absorbance"
    LUMINESCENCE = "luminescence"
    OPTICAL_IMAGING = "optical imaging"

    @staticmethod
    def create(experiment_type: str) -> ExperimentType:
        experiment_type_lower = experiment_type.lower()
        if "fluorescence" in experiment_type_lower or "alpha" in experiment_type_lower:
            return ExperimentType.FLUORESCENCE
        elif "abs" in experiment_type_lower:
            return ExperimentType.ABSORBANCE
        elif "luminescence" in experiment_type_lower:
            return ExperimentType.LUMINESCENCE
        elif "img" in experiment_type_lower:
            return ExperimentType.OPTICAL_IMAGING

        msg = f"Unable to determine experiment type from: '{experiment_type}'"
        raise AllotropeConversionError(msg)

    @property
    def measurement_type(self) -> MeasurementType:
        if self is ExperimentType.FLUORESCENCE:
            return MeasurementType.FLUORESCENCE
        elif self is ExperimentType.ABSORBANCE:
            return MeasurementType.ULTRAVIOLET_ABSORBANCE
        elif self is ExperimentType.LUMINESCENCE:
            return MeasurementType.LUMINESCENCE
        elif self is ExperimentType.OPTICAL_IMAGING:
            return MeasurementType.OPTICAL_IMAGING

    @property
    def device_type(self) -> str:
        if self is ExperimentType.OPTICAL_IMAGING:
            return "imaging detector"
        return f"{self.value} detector"

    @property
    def detection_type(self) -> str | None:
        # TODO(nstender): investigate why we override detection type for these two type but not others,
        # when we have detection type value for all examples.
        if self is ExperimentType.OPTICAL_IMAGING:
            return "optical-imaging"
        elif self is ExperimentType.FLUORESCENCE:
            return str(self.value)
        return None


class Version(Enum):
    V2 = "2.0"
    V3 = "3.0"


@dataclass
class BackgroundInfo:
    experiment_type_value: str

    @property
    def experiment_type(self) -> ExperimentType:
        return ExperimentType.create(self.experiment_type_value)

    @staticmethod
    def create(reader: CsvReader) -> BackgroundInfo:
        line = assert_not_none(
            reader.drop_until_inclusive("^Results? for.(.+) 1"),
            msg="Unable to find background information.",
        )

        experiment_type = assert_not_none(
            re.match("^Results? for.(.+) 1", line),
            msg="Unable to find experiment type from background information section.",
        ).group(1)

        return BackgroundInfo(experiment_type_value=experiment_type)


@dataclass(frozen=True)
class Results:
    barcode: str
    data: dict[str, str]

    @property
    def plate_well_count(self) -> int:
        return len(self.data)

    @classmethod
    def create(cls, reader: CsvReader) -> Results:
        barcode = cls.read_barcode(reader)

        results = assert_not_none(
            reader.pop_csv_block_as_df(
                header=0, index_col=0, empty_pat=EMPTY_STR_OR_CSV_LINE
            ),
            msg="Unable to find results table.",
        )

        for column in results:
            if str(column).startswith("Unnamed"):
                results = results.drop(columns=column)

        return Results(
            barcode=barcode,
            data={
                f"{row}{col}": str(values[col])
                for row, values in results.iterrows()
                for col in results.columns
            },
        )

    @classmethod
    def read_barcode(cls, reader: CsvReader) -> str:
        raise NotImplementedError


@dataclass(frozen=True)
class AnalysisResult:
    analysis_parameter: str
    results: dict[str, str]

    @staticmethod
    def create_results(
        reader: CsvReader, next_section_title: str
    ) -> list[AnalysisResult]:
        section_title = assert_not_none(
            reader.drop_until(f"^Results? for|^{next_section_title}"),
            msg=f"Unable to find Analysis Result or {next_section_title} section.",
        )

        if section_title.startswith(next_section_title):
            return []

        reader.drop_until("^Barcode")

        analysis_results = []
        while reader.match("^Barcode"):
            if analysis_result := AnalysisResult.create(reader):
                analysis_results.append(analysis_result)

        return analysis_results

    @staticmethod
    def create(reader: CsvReader) -> AnalysisResult | None:
        barcode_line = assert_not_none(
            reader.drop_until_inclusive("^Barcode:(.+),"),
            msg="Unable to find background information.",
        )

        analysis_parameter = None
        for element in barcode_line.split(","):
            if "AnalysisParameter" in element:
                analysis_parameter = element.split(":", maxsplit=1)[1]

        analysis_parameter = assert_not_none(
            analysis_parameter,
            msg="Unable to find analysis parameter in Analysis Results section.",
        )

        try:
            results_df = assert_not_none(
                reader.pop_csv_block_as_df(
                    header=0, index_col=0, empty_pat=EMPTY_STR_OR_CSV_LINE
                ),
                msg="Unable to find results table.",
            ).dropna(how="all")
        except AllotropeParsingError:
            logging.warning(
                f"Unable to read analysis result '{analysis_parameter}'. Ignoring"
            )
            return None

        results = {
            f"{row}{col}": str(values[col])
            for row, values in results_df.iterrows()
            for col in results_df.columns
        }

        # if first value is not a valid float value the results are not useful
        if try_float_or_none(results.get("A1")) is None:
            return None

        return AnalysisResult(
            analysis_parameter=analysis_parameter,
            results=results,
        )

    def get_image_feature_result(self, well_position: str) -> float:
        value_name = f"analysis result '{self.analysis_parameter}' at '{well_position}'"
        return try_float(
            assert_not_none(self.results.get(well_position), value_name), value_name
        )


@dataclass(frozen=True)
class MeasurementInfo:
    instrument_serial_number: str
    measurement_time: str
    protocol_signature: str
    measurement_signature: str

    @staticmethod
    def create(
        reader: CsvReader, section_name: str, next_section_title: str
    ) -> MeasurementInfo:
        assert_not_none(
            reader.drop_until_inclusive(f"^{section_name}"),
            msg=f"Unable to find {section_name} section.",
        )

        lines = list(reader.pop_until(f"^{next_section_title}"))
        # Because we may read over multiple sections of data to get all metadata for this section, there may
        # be varying number of columns. Get the max number of columns, so we can read in all sections without error.
        max_num_cols = max(len(line.split(",")) for line in lines)
        df = assert_not_none(
            reader.lines_as_df(lines, index_col=0, names=range(max_num_cols)),
            msg=f"Unable to parser data for {section_name} section.",
        ).T.dropna(how="all")
        df.columns = df.columns.astype(str).str.strip(":")
        data = df_to_series_data(df, index=0)

        return MeasurementInfo(
            instrument_serial_number=data[str, "Instrument Serial Number"],
            measurement_time=data[str, "Measurement Started"],
            protocol_signature=data[str, "Protocol Signature"],
            measurement_signature=data[str, "Measurement Signature"],
        )


@dataclass(frozen=True)
class Platemap:
    data: dict[str, str]

    @staticmethod
    def create(reader: CsvReader) -> Platemap:
        assert_not_none(
            reader.drop_until_inclusive("^Platemap"),
            msg="Unable to find Platemap section.",
        )
        reader.pop_if_match("^Plate")

        data = assert_not_none(
            reader.pop_csv_block_as_df(
                header=0, index_col=0, empty_pat=EMPTY_STR_OR_CSV_LINE
            ),
            msg="Unable to find platemap information.",
        )

        return Platemap(
            data={
                f"{row}{col}": str(values[col])
                for row, values in data.iterrows()
                for col in data.columns
            }
        )

    def get(self, well_position: str) -> str:
        return assert_not_none(
            self.data.get(well_position),
            f"platemap value for well position '{well_position}'",
        )

    def get_sample_role_type(self, well_position: str) -> SampleRoleType | None:
        raw_value = self.get(well_position)
        if raw_value == "-":
            return None

        value = assert_not_none(
            re.match(r"^([A-Z]+)\d+$", raw_value),
            msg=f"Unable to understand platemap value '{raw_value}' for well position '{well_position}'.",
        ).group(1)

        return assert_not_none(
            constants.PLATEMAP_TO_SAMPLE_ROLE_TYPE.get(value),
            msg=f"Unable to find sample role type for well position '{well_position}'.",
        )


@dataclass(frozen=True)
class Channel:
    name: str
    fluorescent_tag: str | None
    transmitted_light: TransmittedLightSetting | None
    excitation_wavelength: float
    illumination: float
    exposure_duration: float
    additional_focus_offset: float

    @staticmethod
    def create(
        name: str,
        excitation_wavelength: str,
        excitation_power: str,
        exposure_time: str,
        additional_focus_offset: str,
    ) -> Channel:
        return Channel(
            name=name,
            fluorescent_tag=None if name == "BRIGHTFIELD" else name,
            transmitted_light=constants.TRANSMITTED_LIGHT_CONVERSION.get(name),
            excitation_wavelength=try_float(
                excitation_wavelength.removesuffix("nm"), "excitation wavelength"
            ),
            illumination=try_float(excitation_power, "excitation power"),
            exposure_duration=try_float(exposure_time, "exposure time"),
            additional_focus_offset=try_float(
                additional_focus_offset, "additional focus offset"
            ),
        )


@dataclass(frozen=True)
class Measurements:
    channels: list[Channel]
    number_of_averages: float | None
    detector_distance: float | None
    scan_position: ScanPositionSettingPlateReader | None
    emission_wavelength: float | None
    excitation_wavelength: float | None
    focus_height: float | None

    @staticmethod
    def create(
        reader: CsvReader, section_title: str, next_section_title: str
    ) -> Measurements:
        assert_not_none(
            reader.drop_until_inclusive(f"^{section_title}"),
            msg=f"Unable to find {section_title} section.",
        )

        lines = list(reader.pop_until(f"^{next_section_title}"))
        # Because we may read over multiple sections of data to get all metadata for this section, there may
        # be varying number of columns. Get the max number of columns, so we can read in all sections without error.
        max_num_cols = max(len(line.split(",")) for line in lines)
        df = assert_not_none(
            reader.lines_as_df(lines, index_col=0, names=range(max_num_cols)),
            msg=f"Unable to parser data for {section_title} section.",
        ).T.dropna(how="all")
        df.columns = df.columns.astype(str).str.lower()
        data = df_to_series_data(df, index=0)

        scan_position = data.get(str, "excitation / emission")
        excitation_wavelength = data.get(str, "excitation wavelength [nm]")

        return Measurements(
            channels=Measurements.create_channels(data),
            number_of_averages=data.get(float, "number of flashes"),
            detector_distance=data.get(
                float, "distance between plate and detector [mm]"
            ),
            scan_position=(
                None
                if scan_position is None
                else assert_not_none(
                    constants.SCAN_POSITION_CONVERSION.get(scan_position),
                    msg=f"'{scan_position}' is not a valid scan position, expected TOP or BOTTOM.",
                )
            ),
            emission_wavelength=data.get(float, "emission wavelength [nm]"),
            excitation_wavelength=(
                None
                if excitation_wavelength is None
                else try_float_or_none(excitation_wavelength.removesuffix("nm"))
            ),
            focus_height=data.get(float, "focus height [µm]"),
        )

    @staticmethod
    def create_channels(data: SeriesData) -> list[Channel]:
        # Get all channel keys
        values = {
            key: data.series.get(key)
            for key in [
                "channel",
                "excitation wavelength [nm]",
                "excitation power [%]",
                "exposure time [ms]",
                "additional focus offset [mm]",
            ]
        }
        if values["channel"] is None:
            return []

        # Convert series values (multiple Channels == pd.Series, single Channel == str) to lists.
        channel_values: dict[str, list[str]] = {}
        if isinstance(values["channel"], str):
            for key, value in values.items():
                if not isinstance(value, str):
                    raise AllotropeConversionError(constants.CHANNEL_COLUMNS_ERROR)
                channel_values[key] = [value]
        elif isinstance(values["channel"], pd.Series):
            for key, value in values.items():
                if not isinstance(value, pd.Series):
                    raise AllotropeConversionError(constants.CHANNEL_COLUMNS_ERROR)
                channel_values[key] = list(value.values)
                if not len(channel_values["channel"]) == len(channel_values[key]):
                    raise AllotropeConversionError(constants.CHANNEL_COLUMNS_ERROR)

        return [
            Channel.create(
                name=channel_values["channel"][i],
                excitation_wavelength=channel_values["excitation wavelength [nm]"][i],
                excitation_power=channel_values["excitation power [%]"][i],
                exposure_time=channel_values["exposure time [ms]"][i],
                additional_focus_offset=channel_values["additional focus offset [mm]"][
                    i
                ],
            )
            for i in range(len(channel_values["channel"]))
        ]


@dataclass(frozen=True)
class Data:
    version: str
    background_info: BackgroundInfo
    results: Results
    analysis_results: list[AnalysisResult]
    measurement_info: MeasurementInfo
    platemap: Platemap
    measurements: Measurements

    def iter_wells(self) -> Iterator[tuple[str, str]]:
        yield from self.results.data.items()

    def get_sample_identifier(self, well_position: str) -> str:
        platemap_value = self.platemap.get(well_position)
        return (
            f"{self.results.barcode}_{well_position}"
            if platemap_value in ("-", None)
            else platemap_value
        )


def create_metadata(data: Data, file_path: str) -> Metadata:
    return Metadata(
        file_name=Path(file_path).name,
        unc_path=file_path,
        software_name=constants.SOFTWARE_NAME,
        software_version=data.version,
        device_identifier=constants.DEVICE_IDENTIFIER,
        model_number=constants.MODEL_NUMBER,
        product_manufacturer=constants.PRODUCT_MANUFACTURER,
        equipment_serial_number=data.measurement_info.instrument_serial_number,
    )


def _create_measurement(data: Data, well_position: str, well_value: str) -> Measurement:
    experiment_type = data.background_info.experiment_type

    if experiment_type is ExperimentType.ABSORBANCE:
        assert_not_none(
            data.measurements.number_of_averages,
            msg="Unable to find number of averages",
        )

    detector_wavelength_setting = (
        data.measurements.emission_wavelength
        if experiment_type is ExperimentType.FLUORESCENCE
        else (
            data.measurements.excitation_wavelength
            if experiment_type is ExperimentType.ABSORBANCE
            else None
        )
    )
    measurement_value = try_float(well_value, f"result well at '{well_position}'")

    return Measurement(
        type_=experiment_type.measurement_type,
        identifier=random_uuid_str(),
        device_type=experiment_type.device_type,
        detection_type=experiment_type.detection_type
        or data.background_info.experiment_type_value,
        sample_identifier=data.get_sample_identifier(well_position),
        location_identifier=well_position,
        well_plate_identifier=data.results.barcode,
        sample_role_type=data.platemap.get_sample_role_type(well_position),
        number_of_averages=data.measurements.number_of_averages,
        detector_distance_setting=data.measurements.detector_distance,
        scan_position_setting=data.measurements.scan_position,
        detector_wavelength_setting=detector_wavelength_setting,
        excitation_wavelength_setting=data.measurements.excitation_wavelength,
        fluorescence=measurement_value
        if experiment_type is ExperimentType.FLUORESCENCE
        else None,
        absorbance=measurement_value
        if experiment_type is ExperimentType.ABSORBANCE
        else None,
        luminescence=measurement_value
        if experiment_type is ExperimentType.LUMINESCENCE
        else None,
    )


def _create_optical_measurement(
    data: Data, well_position: str, channel: Channel
) -> Measurement:
    return Measurement(
        type_=data.background_info.experiment_type.measurement_type,
        identifier=random_uuid_str(),
        device_type=data.background_info.experiment_type.device_type,
        detection_type=data.background_info.experiment_type.detection_type
        or data.background_info.experiment_type_value,
        sample_identifier=data.get_sample_identifier(well_position),
        location_identifier=well_position,
        well_plate_identifier=data.results.barcode,
        sample_role_type=data.platemap.get_sample_role_type(well_position),
        detector_distance_setting=data.measurements.focus_height,
        excitation_wavelength_setting=channel.excitation_wavelength,
        magnification_setting=constants.MAGNIFICATION_SETTING,
        exposure_duration_setting=channel.exposure_duration,
        illumination_setting=channel.illumination,
        illumination_setting_unit="%",
        transmitted_light_setting=channel.transmitted_light,
        fluorescent_tag_setting=channel.fluorescent_tag,
    )


def _create_measurements(
    data: Data, well_position: str, well_value: str
) -> list[Measurement]:
    if data.background_info.experiment_type is ExperimentType.OPTICAL_IMAGING:
        return [
            _create_optical_measurement(data, well_position, channel)
            for channel in data.measurements.channels
        ]
    else:
        return [_create_measurement(data, well_position, well_value)]


def create_measurement_groups(data: Data) -> list[MeasurementGroup]:
    return [
        MeasurementGroup(
            measurement_time=data.measurement_info.measurement_time,
            plate_well_count=data.results.plate_well_count,
            experiment_type=data.background_info.experiment_type_value,
            analytical_method_identifier=data.measurement_info.protocol_signature,
            experimental_data_identifier=data.measurement_info.measurement_signature,
            measurements=_create_measurements(data, well_position, well_value),
            processed_data=ProcessedData(
                features=[
                    ImageFeature(
                        identifier=random_uuid_str(),
                        feature=analysis_result.analysis_parameter,
                        result=analysis_result.get_image_feature_result(well_position),
                        data_sources=[
                            DataSource(
                                identifier=well_value,
                                # TODO(nstender): I'm not sure what this original comment means, investigate.
                                feature="image feature",  # wait for actual value
                            )
                        ],
                    )
                    for analysis_result in data.analysis_results
                ]
            ),
            images=[
                ImageSource(identifier=well_value),
            ]
            if data.background_info.experiment_type is ExperimentType.OPTICAL_IMAGING
            else None,
        )
        for well_position, well_value in data.iter_wells()
    ]



# plate (repeated for N plates measured)
#     Plate information
#     Background information (optional)
#     Calculated results (optional)
#     Measured results (repeated for N measurements performed per plate)
# Basic assay information
# Protocol information
# Plate type
# Platemap (repeated for N plates measured)
# Calculations
# Auto export parameters
# Operations
# Labels
#     Filters (dependent on detection modality)
#     Mirrors (dependent on detection modality)
# Instrument
from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from re import search

import numpy as np
import pandas as pd

from allotropy.allotrope.models.shared.components.plate_reader import SampleRoleType
from allotropy.allotrope.models.shared.definitions.units import UNITLESS
from allotropy.allotrope.schema_mappers.adm.plate_reader.rec._2024._06.plate_reader import (
    CalculatedDataItem,
    DataSource,
    Measurement,
    MeasurementGroup,
    MeasurementType,
    Metadata,
    ScanPositionSettingPlateReader,
)
from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.perkin_elmer_envision import constants
from allotropy.parsers.utils.pandas import (
    df_to_series_data,
    map_rows,
    SeriesData,
)
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import (
    assert_not_none,
    num_to_chars,
)


class ReadType(Enum):
    ABSORBANCE = "Absorbance"
    FLUORESCENCE = "Fluorescence"
    LUMINESCENCE = "Luminescence"

    @property
    def measurement_type(self) -> MeasurementType:
        if self is ReadType.ABSORBANCE:
            return MeasurementType.ULTRAVIOLET_ABSORBANCE
        elif self is ReadType.FLUORESCENCE:
            return MeasurementType.FLUORESCENCE
        elif self is ReadType.LUMINESCENCE:
            return MeasurementType.LUMINESCENCE

    @property
    def device_type(self) -> str:
        return f"{self.value.lower()} detector"


@dataclass(frozen=True)
class PlateInfo:
    number: str
    barcode: str
    measurement_time: str | None
    measured_height: float | None
    chamber_temperature_at_start: float | None

    @staticmethod
    def get_series(reader: CsvReader) -> SeriesData:
        assert_not_none(
            reader.pop_if_match("^Plate information"),
            msg="Unable to find expected plate information",
        )

        return df_to_series_data(
            assert_not_none(
                reader.pop_csv_block_as_df(header=0),
                "Plate information CSV block",
            ).replace(np.nan, None)
        )

    @staticmethod
    def get_plate_number_and_barcode(data: SeriesData) -> tuple[str, str]:
        plate_number = data[str, "Plate"]
        raw_barcode = data.get(str, "Barcode")
        barcode = (raw_barcode or '=""').removeprefix('="').removesuffix('"')
        barcode = barcode or f"Plate {plate_number}"
        return plate_number, barcode


@dataclass(frozen=True)
class CalculatedPlateInfo(PlateInfo):
    formula: str
    name: str

    @staticmethod
    def create(data: SeriesData) -> CalculatedPlateInfo:
        formula = data[str, "Formula"]

        name = assert_not_none(
            search(r"^([^=]*)=", formula),
            msg="Unable to find expected formula name for calculated results section.",
        ).group(1)

        plate_number, barcode = PlateInfo.get_plate_number_and_barcode(data)

        return CalculatedPlateInfo(
            number=plate_number,
            barcode=barcode,
            measurement_time=data.get(str, "Measurement date"),
            measured_height=data.get(float, "Measured height"),
            chamber_temperature_at_start=data.get(
                float, "Chamber temperature at start"
            ),
            formula=formula,
            name=name.strip(),
        )


@dataclass(frozen=True)
class ResultPlateInfo(PlateInfo):
    label: str
    measinfo: str
    emission_filter_id: str

    @staticmethod
    def create(data: SeriesData) -> ResultPlateInfo | None:
        label = data.get(str, "Label")
        if label is None:
            return None

        measinfo = data.get(str, "Measinfo")
        if measinfo is None:
            return None

        plate_number, barcode = PlateInfo.get_plate_number_and_barcode(data)

        return ResultPlateInfo(
            plate_number,
            barcode,
            data.get(str, "Measurement date"),
            data.get(float, "Measured height"),
            data.get(float, "Chamber temperature at start"),
            label=label,
            measinfo=measinfo,
            emission_filter_id=assert_not_none(
                search("De=(...)", measinfo),
                msg=f"Unable to find emission filter ID for Plate {barcode}.",
            ).group(1),
        )

    def match(self, background_info: BackgroundInfo) -> bool:
        return all(
            [
                background_info.plate_num == self.number,
                background_info.label in self.label,
                background_info.measinfo == self.measinfo,
            ]
        )


@dataclass
class BackgroundInfo:
    plate_num: str
    label: str
    measinfo: str

    @staticmethod
    def create(data: SeriesData) -> BackgroundInfo:
        return BackgroundInfo(
            plate_num=data[str, "Plate"],
            label=data[str, "Label"],
            measinfo=data[str, "MeasInfo"],
        )


def create_background_infos(reader: CsvReader) -> list[BackgroundInfo]:
    title = reader.pop_if_match("^Background information")
    if title is None:
        return []

    data_frame = assert_not_none(
        reader.pop_csv_block_as_df(header=0),
        "background information",
    )
    return map_rows(data_frame, BackgroundInfo.create)


@dataclass
class CalculatedResult:
    uuid: str
    col: str
    row: str
    value: float


def create_calculated_results(reader: CsvReader) -> list[CalculatedResult]:
    # Calculated results may or may not have a title
    reader.pop_if_match("^Calculated results")

    data_frame = assert_not_none(
        reader.pop_csv_block_as_df(),
        "results data",
    )
    series = (
        data_frame.drop(0, axis="index").drop(0, axis="columns")
        if data_frame.iloc[1, 0] == "A"
        else data_frame
    )
    rows, cols = series.shape
    series.index = pd.Index([num_to_chars(i) for i in range(rows)])
    series.columns = pd.Index([str(i) for i in range(1, cols + 1)])

    return [
        CalculatedResult(
            uuid=random_uuid_str(),
            col=col,
            row=row,
            value=series.loc[col, row],
        )
        for col, row in series.stack().index
    ]


@dataclass
class Result:
    uuid: str
    col: str
    row: str
    value: int


def create_results(reader: CsvReader) -> list[Result]:
    # Results may or may not have a title
    reader.pop_if_match("^Results")

    data_frame = assert_not_none(
        reader.pop_csv_block_as_df(),
        "reader data",
    )
    series = (
        data_frame.drop(0, axis="index").drop(0, axis="columns")
        if data_frame.iloc[1, 0] == "A"
        else data_frame
    )
    rows, cols = series.shape
    series.index = pd.Index([num_to_chars(i) for i in range(rows)])
    series.columns = pd.Index([str(i) for i in range(1, cols + 1)])

    return [
        Result(
            uuid=random_uuid_str(),
            col=col,
            row=row,
            value=int(series.loc[col, row]),
        )
        for col, row in series.stack().index
    ]


@dataclass
class Plate:
    plate_info: PlateInfo
    background_infos: list[BackgroundInfo]


@dataclass
class ResultPlate(Plate):
    plate_info: ResultPlateInfo
    results: list[Result]


@dataclass
class CalculatedPlate(Plate):
    plate_info: CalculatedPlateInfo
    results: list[CalculatedResult]

    def get_source_results(self, plate_list: PlateList) -> Iterator[list[Result]]:
        if not self.background_infos:
            msg = f"Unable to collect result plates for calculated data, there is no background information for plate {self.plate_info.number}"
            raise AllotropeConversionError(msg)

        source_results = [
            assert_not_none(
                plate_list.get_result_plate(background_info),
                msg=f"Unable to find result plate {background_info.label}.",
            ).results
            for background_info in self.background_infos
        ]

        if not all(len(source_results[0]) == len(sr) for sr in source_results):
            msg = f"Unable to collect result plates for calculated data, expected all result plates to have the same number of results, get: {[len(sr) for sr in source_results]}."
            raise AllotropeConversionError(msg)

        yield from [list(tup) for tup in zip(*source_results, strict=True)]

    def get_result_and_sources(
        self, plate_list: PlateList
    ) -> Iterator[tuple[CalculatedResult, list[Result]]]:
        yield from zip(self.results, self.get_source_results(plate_list), strict=True)


def create_plate(reader: CsvReader) -> ResultPlate | CalculatedPlate:
    series = PlateInfo.get_series(reader)
    if result_plate_info := ResultPlateInfo.create(series):
        return ResultPlate(
            plate_info=result_plate_info,
            background_infos=create_background_infos(reader),
            results=create_results(reader),
        )
    else:
        return CalculatedPlate(
            plate_info=CalculatedPlateInfo.create(series),
            background_infos=create_background_infos(reader),
            results=create_calculated_results(reader),
        )


@dataclass
class PlateList:
    results: list[ResultPlate]
    calculated: list[CalculatedPlate]

    @staticmethod
    def create(reader: CsvReader) -> PlateList:
        results: list[ResultPlate] = []
        calculated: list[CalculatedPlate] = []

        while reader.match("^Plate information"):
            plate = create_plate(reader)
            if isinstance(plate, CalculatedPlate):
                calculated.append(plate)
            else:
                results.append(plate)
        return PlateList(results, calculated)

    def get_result_plate(self, background_info: BackgroundInfo) -> ResultPlate | None:
        for plate in self.results:
            if plate.plate_info.match(background_info):
                return plate
        return None

    def get_measurement_time(self) -> str:
        try:
            return min(
                [
                    plate.plate_info.measurement_time
                    for plate in self.results
                    if plate.plate_info.measurement_time
                ]
            )
        except ValueError as err:
            msg = "Unable to determine the measurement time."
            raise AllotropeConversionError(msg) from err


@dataclass(frozen=True)
class BasicAssayInfo:
    protocol_id: str | None
    assay_id: str | None

    @staticmethod
    def create(reader: CsvReader) -> BasicAssayInfo:
        reader.drop_until_inclusive("^Basic assay information")
        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(index_col=0),
            "Basic assay information",
        ).T
        data_frame.columns = (
            data_frame.columns.astype(str).str.replace(":", "").str.strip()
        )
        data = df_to_series_data(data_frame)
        return BasicAssayInfo(data.get(str, "Protocol ID"), data.get(str, "Assay ID"))


@dataclass(frozen=True)
class PlateType:
    number_of_wells: float

    @staticmethod
    def create(reader: CsvReader) -> PlateType:
        reader.drop_until_inclusive("^Plate type")
        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(index_col=0), "Plate type"
        )
        data = df_to_series_data(data_frame.T)
        return PlateType(
            number_of_wells=data[float, "Number of the wells in the plate"]
        )


def get_sample_role_type(encoding: str) -> SampleRoleType:
    # BL        blank               blank_role
    # CTL       control             control_sample_role
    # LB        lance_blank         blank_role
    # LC        lance_crosstalk     control_sample_role
    # LH        lance_high          control_sample_role
    # S         pl_sample           sample_role
    # STD       standard            standard_sample_role
    # -         unknown             unknown_sample_role
    # UNK       unknown             unknown_sample_role
    # ZH        z_high              control_sample_role
    # ZL        z_low               control_sample_role
    sample_role_type_map = {
        "BL": SampleRoleType.blank_role,
        "CTL": SampleRoleType.control_sample_role,
        "LB": SampleRoleType.blank_role,
        "LC": SampleRoleType.control_sample_role,
        "LH": SampleRoleType.control_sample_role,
        "STD": SampleRoleType.standard_sample_role,
        "S": SampleRoleType.sample_role,
        "-": SampleRoleType.unknown_sample_role,
        "UNK": SampleRoleType.unknown_sample_role,
        "ZH": SampleRoleType.control_sample_role,
        "ZL": SampleRoleType.control_sample_role,
    }
    for pattern, value in sample_role_type_map.items():
        if encoding.startswith(pattern):
            return value

    msg = f"Unable to determine sample role type of plate map encoding {encoding}; expected to start with one of {sorted(sample_role_type_map.keys())}."
    raise AllotropeConversionError(msg)


@dataclass(frozen=True)
class PlateMap:
    plate_n: str
    group_n: str
    sample_role_type_mapping: dict[str, dict[str, SampleRoleType]]

    @staticmethod
    def create(reader: CsvReader) -> PlateMap | None:
        if not reader.current_line_exists() or reader.match("^Calculations"):
            return None

        *_, plate_n = assert_not_none(reader.pop(), "Platemap number").split(",")
        *_, group_n = assert_not_none(reader.pop(), "Platemap group").split(",")

        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(),
            "Platemap data",
        ).replace(" ", "", regex=True)

        reader.pop_data()  # drop type specification
        reader.drop_empty()

        series = (
            data_frame.drop(0, axis="index").drop(0, axis="columns")
            if data_frame.iloc[1, 0] == "A"
            else data_frame
        )
        rows, cols = series.shape
        series.index = pd.Index([num_to_chars(i) for i in range(rows)])
        series.columns = pd.Index([str(i) for i in range(1, cols + 1)])

        sample_role_type_mapping: dict[str, dict[str, SampleRoleType]] = {}
        for row, row_data in series.replace([np.nan, "''"], None).to_dict().items():
            col_mapping: dict[str, SampleRoleType] = {}
            for col, value in row_data.items():
                if value:
                    if role_type := get_sample_role_type(str(value)):
                        col_mapping[str(col)] = role_type
            if col_mapping:
                sample_role_type_mapping[str(row)] = col_mapping

        return PlateMap(plate_n, group_n, sample_role_type_mapping)

    def get_sample_role_type(self, col: str, row: str) -> SampleRoleType:
        try:
            return self.sample_role_type_mapping[row][col]
        except KeyError as e:
            msg = (
                f"Invalid plate map location for plate map {self.plate_n}: {col} {row}."
            )
            raise AllotropeConversionError(msg) from e


def create_plate_maps(reader: CsvReader) -> dict[str, PlateMap]:
    assert_not_none(
        reader.drop_until("^Platemap"),
        msg="No 'Platemap' section found.",
    )

    reader.pop()  # remove title

    maps: dict[str, PlateMap] = {}
    while _map := PlateMap.create(reader):
        maps[_map.plate_n] = _map
    return maps


@dataclass(frozen=True)
class Filter:
    name: str
    wavelength: float
    bandwidth: float | None = None

    @staticmethod
    def create(reader: CsvReader) -> Filter | None:
        if not reader.current_line_exists() or reader.match(
            "(^Mirror modules)|(^Instrument:)|(^Aperture:)"
        ):
            return None

        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(index_col=0),
            "Filter information",
        )
        data = df_to_series_data(data_frame.T)
        name = str(data.series.index[0])
        description = data[str, "Description"]

        if search_result := search("Longpass=(\\d+)nm", description):
            return Filter(name, wavelength=float(search_result.group(1)))
        return Filter(
            name,
            wavelength=float(
                assert_not_none(
                    search("CWL=(\\d+)nm", description),
                    msg=f"Unable to find wavelength for filter {name}.",
                ).group(1)
            ),
            bandwidth=float(
                assert_not_none(
                    search("BW=(\\d+)nm", description),
                    msg=f"Unable to find bandwidth for filter {name}.",
                ).group(1)
            ),
        )


def create_filters(reader: CsvReader) -> dict[str, Filter]:
    reader.drop_until("(^Filters:)|^Instrument:")

    if reader.match("^Instrument"):
        return {}

    reader.pop()  # remove title

    filters = {}
    while _filter := Filter.create(reader):
        filters[_filter.name] = _filter
    return filters


@dataclass(frozen=True)
class Labels:
    label: str
    excitation_filter: Filter | None
    emission_filters: dict[str, Filter | None]
    scan_position_setting: ScanPositionSettingPlateReader | None = None
    number_of_flashes: float | None = None
    detector_gain_setting: str | None = None

    @staticmethod
    def create(reader: CsvReader) -> Labels:
        reader.drop_until_inclusive("^Labels")
        data_frame = assert_not_none(
            reader.pop_csv_block_as_df(index_col=0),
            "Labels",
        )
        data = df_to_series_data(data_frame.T.replace(np.nan, None))
        filters = create_filters(reader)
        filter_position_map = {
            "Bottom": ScanPositionSettingPlateReader.bottom_scan_position__plate_reader_,
            "Top": ScanPositionSettingPlateReader.top_scan_position__plate_reader_,
        }

        return Labels(
            label=data.series.index[0],
            excitation_filter=filters.get(data.get(str, "Exc. filter", NOT_APPLICABLE)),
            emission_filters={
                "1st": filters.get(
                    data.get(str, "Ems. filter", NOT_APPLICABLE),
                ),
                "2nd": filters.get(data.get(str, "2nd ems. filter", NOT_APPLICABLE)),
            },
            scan_position_setting=filter_position_map.get(
                data.get(str, "Using of emission filter", NOT_APPLICABLE)
            ),
            number_of_flashes=data.get(float, "Number of flashes"),
            detector_gain_setting=data.get(str, "Reference AD gain"),
        )

    def get_emission_filter(self, id_val: str) -> Filter | None:
        return self.emission_filters.get(id_val)

    def get_read_type(self) -> ReadType:
        patterns = {
            "ABS": ReadType.ABSORBANCE,
            "Absorbance": ReadType.ABSORBANCE,
            "LUM": ReadType.LUMINESCENCE,
            "Luminescence": ReadType.LUMINESCENCE,
            "Fluorescence": ReadType.FLUORESCENCE,
        }
        read_types = {
            read_type for key, read_type in patterns.items() if key in self.label
        }

        if len(read_types) > 1:
            msg = f"Unable to determine unique read type from label: '{self.label}'"
            raise AllotropeConversionError(msg)
        if len(read_types) == 1:
            return read_types.pop()
        # TODO check if this is correct, this is the original behavior
        return ReadType.FLUORESCENCE


@dataclass(frozen=True)
class Instrument:
    serial_number: str
    nickname: str

    @staticmethod
    def create(reader: CsvReader) -> Instrument:
        assert_not_none(
            reader.drop_until("^Instrument"),
            msg="No 'Instrument' section found.",
        )

        reader.pop()  # remove title

        *_, serial_number = assert_not_none(reader.pop(), "serial number").split(",")
        *_, nickname = assert_not_none(reader.pop(), "nickname").split(",")

        return Instrument(serial_number, nickname)


@dataclass(frozen=True)
class Software:
    software_name: str
    software_version: str

    @staticmethod
    def create(reader: CsvReader) -> Software:
        exported_with_regex = "Exported with (.+) version (.+)"
        assert_not_none(
            reader.drop_until(exported_with_regex),
            msg="Unable to find software information; no 'Exported with' section found.",
        )

        search_result = assert_not_none(
            search(
                exported_with_regex,
                assert_not_none(
                    reader.pop(),
                    "software information",
                ),
            )
        )

        return Software(
            software_name=search_result.group(1),
            software_version=search_result.group(2),
        )


@dataclass(frozen=True)
class Data:
    software: Software
    plate_list: PlateList
    basic_assay_info: BasicAssayInfo
    number_of_wells: float
    plate_maps: dict[str, PlateMap]
    labels: Labels
    instrument: Instrument

    @staticmethod
    def create(reader: CsvReader) -> Data:
        return Data(
            plate_list=PlateList.create(reader),
            basic_assay_info=BasicAssayInfo.create(reader),
            number_of_wells=PlateType.create(reader).number_of_wells,
            plate_maps=create_plate_maps(reader),
            labels=Labels.create(reader),
            instrument=Instrument.create(reader),
            software=Software.create(reader),
        )


def create_metadata(
    software: Software, instrument: Instrument, file_path: str
) -> Metadata:
    path = Path(file_path)
    return Metadata(
        file_name=path.name,
        unc_path=file_path,
        asm_file_identifier=path.with_suffix(".json").name,
        software_name=software.software_name,
        software_version=software.software_version,
        model_number=constants.MODEL_NUMBER,
        data_system_instance_id=NOT_APPLICABLE,
        equipment_serial_number=instrument.serial_number,
        device_identifier=instrument.nickname,
    )


def _create_measurement(
    plate_info: ResultPlateInfo,
    result: Result,
    plate_maps: dict[str, PlateMap],
    labels: Labels,
    read_type: ReadType,
) -> Measurement:
    try:
        p_map = plate_maps[plate_info.number]
    except KeyError as e:
        msg = f"Unable to find plate map for Plate {plate_info.barcode}."
        raise AllotropeConversionError(msg) from e
    plate_barcode = plate_info.barcode
    well_location = f"{result.col}{result.row}"
    ex_filter = labels.excitation_filter
    em_filter = labels.get_emission_filter(plate_info.emission_filter_id)
    return Measurement(
        type_=read_type.measurement_type,
        device_type=read_type.device_type,
        identifier=result.uuid,
        sample_identifier=f"{plate_barcode} {well_location}",
        well_plate_identifier=plate_barcode,
        location_identifier=well_location,
        sample_role_type=p_map.get_sample_role_type(result.col, result.row),
        compartment_temperature=plate_info.chamber_temperature_at_start,
        absorbance=result.value if read_type is ReadType.ABSORBANCE else None,
        fluorescence=result.value if read_type is ReadType.FLUORESCENCE else None,
        luminescence=result.value if read_type is ReadType.LUMINESCENCE else None,
        detector_distance_setting=plate_info.measured_height,
        number_of_averages=labels.number_of_flashes,
        detector_gain_setting=labels.detector_gain_setting,
        scan_position_setting=labels.scan_position_setting,
        detector_wavelength_setting=em_filter.wavelength if em_filter else None,
        detector_bandwidth_setting=em_filter.bandwidth if em_filter else None,
        excitation_wavelength_setting=ex_filter.wavelength if ex_filter else None,
        excitation_bandwidth_setting=ex_filter.bandwidth if ex_filter else None,
    )


def create_measurement_groups(data: Data) -> list[MeasurementGroup]:
    read_type = data.labels.get_read_type()
    well_loc_measurements = defaultdict(list)
    for plate in data.plate_list.results:
        for result in plate.results:
            measurement = _create_measurement(
                plate.plate_info,
                result,
                data.plate_maps,
                data.labels,
                read_type,
            )
            well_loc_measurements[
                (plate.plate_info.number, measurement.location_identifier)
            ].append(measurement)

    measurement_time = data.plate_list.get_measurement_time()
    return [
        MeasurementGroup(
            measurement_time=measurement_time,
            plate_well_count=data.number_of_wells,
            analytical_method_identifier=data.basic_assay_info.protocol_id,
            experimental_data_identifier=data.basic_assay_info.assay_id,
            measurements=well_loc_measurements[well_location],
        )
        for well_location in sorted(
            well_loc_measurements.keys(),
            key=lambda key: (key[0], key[1][0], int(key[1][1:])),
        )
    ]


def create_calculated_data(
    plate_list: PlateList, read_type: ReadType
) -> list[CalculatedDataItem]:
    return [
        CalculatedDataItem(
            identifier=calculated_result.uuid,
            name=calculated_plate.plate_info.name,
            description=calculated_plate.plate_info.formula,
            value=calculated_result.value,
            unit=UNITLESS,
            data_sources=[
                DataSource(
                    identifier=source_result.uuid,
                    feature=read_type.value,
                )
                for source_result in source_results
            ],
        )
        for calculated_plate in plate_list.calculated
        for calculated_result, source_results in calculated_plate.get_result_and_sources(
            plate_list
        )
    ]



from io import StringIO

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers import lines_reader
from allotropy.parsers.lines_reader import CsvReader
from allotropy.parsers.utils.pandas import read_csv


class Nanodrop8000Reader:
    SUPPORTED_EXTENSIONS = "txt"
    data: pd.DataFrame

    def __init__(self, named_file_contents: NamedFileContents):
        all_lines = lines_reader.read_to_lines(named_file_contents)
        reader = CsvReader(all_lines)
        lines = reader.pop_csv_block_as_lines()
        self.data = read_csv(
            StringIO("\n".join(lines)),
            sep="\t",
            dtype={"Plate ID": str, "Sample ID": str},
            # Prevent pandas from rounding decimal values, at the cost of some speed.
            float_precision="round_trip",
        )
        self.data.columns = self.data.columns.astype(str).str.strip().str.lower()



import numpy as np

from allotropy.allotrope.models.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    Model,
)
from allotropy.allotrope.schema_mappers.adm.solution_analyzer.rec._2024._09.solution_analyzer import (
    Data,
    Mapper,
)
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.novabio_flex2.novabio_flex2_structure import (
    create_measurement_groups,
    create_metadata,
    SampleList,
    Title,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.pandas import read_csv
from allotropy.parsers.vendor_parser import VendorParser


class NovaBioFlexParser(VendorParser[Data, Model]):
    DISPLAY_NAME = "NovaBio Flex2"
    RELEASE_STATE = ReleaseState.RECOMMENDED
    SUPPORTED_EXTENSIONS = "csv"
    SCHEMA_MAPPER = Mapper

    def create_data(self, named_file_contents: NamedFileContents) -> Data:
        # NOTE: calling parse_dates and removing NaN clears empty rows.
        data = read_csv(
            named_file_contents.contents, parse_dates=["Date & Time"]
        ).replace(np.nan, None)
        title = Title.create(named_file_contents.original_file_path)
        sample_list = SampleList.create(data)
        return Data(
            create_metadata(title, named_file_contents.original_file_path),
            create_measurement_groups(title, sample_list),
        )



from __future__ import annotations

from dataclasses import dataclass

from allotropy.exceptions import AllotropeConversionError
from allotropy.parsers.cytiva_unicorn.reader.unicorn_zip_handler import (
    UnicornZipHandler,
)
from allotropy.parsers.utils.strict_xml_element import (
    StrictXmlElement,
)
from allotropy.parsers.utils.uuids import random_uuid_str


@dataclass
class StaticDocs:
    chromatography_serial_num: str
    column_inner_diameter: float
    chromatography_chemistry_type: str
    chromatography_particle_size: float
    injection_identifier: str
    injection_time: str
    autosampler_injection_volume_setting: float
    sample_identifier: str
    batch_identifier: str

    @classmethod
    def create(
        cls,
        handler: UnicornZipHandler,
        curve: StrictXmlElement,
        results: StrictXmlElement,
    ) -> StaticDocs:
        column_type_data = handler.get_column_type_data()
        inj_result = cls.__filter_result_criteria(results, keyword="Sample volume")
        sample_result = cls.__filter_result_criteria(results, keyword="Sample_ID")

        return StaticDocs(
            chromatography_serial_num=column_type_data.recursive_find(
                ["ColumnType", "Hardware", "ArticleNumber"]
            ).get_text(),
            column_inner_diameter=column_type_data.recursive_find(
                ["ColumnType", "Hardware", "Diameter"]
            ).get_float("column inner diameter"),
            chromatography_chemistry_type=column_type_data.recursive_find(
                ["ColumnType", "Media", "TechniqueName"]
            ).get_text(),
            chromatography_particle_size=column_type_data.recursive_find(
                ["ColumnType", "Media", "AverageParticleDiameter"]
            ).get_float("chromatography particle size"),
            injection_identifier=random_uuid_str(),
            injection_time=curve.find("MethodStartTime").get_text(),
            autosampler_injection_volume_setting=inj_result.find("Keyword2").get_float(
                "autosampler injection volume setting"
            ),
            sample_identifier=sample_result.find("Keyword2").get_text(),
            batch_identifier=results.find("BatchId").get_text(),
        )

    @classmethod
    def __filter_result_criteria(
        cls, results: StrictXmlElement, keyword: str
    ) -> StrictXmlElement:
        for result_criteria in results.find("ResultSearchCriterias").findall(
            "ResultSearchCriteria"
        ):
            if result_criteria.find("Keyword1").get_text() == keyword:
                return result_criteria
        msg = f"Unable to find result criteria with keyword 1 '{keyword}'"
        raise AllotropeConversionError(msg)



MODEL_NUMBER = "Flex2"
SOFTWARE_NAME = "NovaBio Flex"
DEVICE_TYPE = "solution-analyzer"
PRODUCT_MANUFACTURER = "Nova Biomedical"

FILENAME_REGEX = r"^SampleResults(?P<device_identifier>.*)?(?P<processing_time>\d{4}-\d{2}-\d{2}_\d{6})\.csv$"

INVALID_FILENAME_MESSAGE = (
    "{} is not valid. File name is expected to have format of SampleResultsYYYY-MM-DD_HHMMSS.csv "
    "or SampleResults<Analyzer ID>YYYY-MM-DD_HHMMSS.csv where <Analyzer ID> is defined in Settings"
)

ANALYTE_MAPPINGS = {
    "NH4+": {
        "name": "ammonium",
        "unit": "mmol/L",
    },
    "HCO3": {
        "name": "bicarbonate",
        "unit": "mmol/L",
    },
    "Ca++": {
        "name": "calcium",
        "unit": "mmol/L",
    },
    "Gluc": {
        "name": "glucose",
        "unit": "g/L",
    },
    "Glu": {
        "name": "glutamate",
        "unit": "mmol/L",
    },
    "Gln": {
        "name": "glutamine",
        "unit": "mmol/L",
    },
    "Lac": {
        "name": "lactate",
        "unit": "g/L",
    },
    "K+": {
        "name": "potassium",
        "unit": "mmol/L",
    },
    "Na+": {
        "name": "sodium",
        "unit": "mmol/L",
    },
}

DETECTION_PROPERTY_MAPPING = {
    "metabolite-detection": ["analytes"],
    "blood-gas-detection": [
        "po2",
        "pco2",
        "carbon_dioxide_saturation",
        "oxygen_saturation",
    ],
    "ph-detection": [
        "ph",
        "temperature",
    ],
    "osmolality-detection": ["osmolality"],
    "cell-counting": [
        "viability",
        "total_cell_density",
        "viable_cell_density",
        "average_live_cell_diameter",
        "total_cell_count",
        "viable_cell_count",
        "cell_type_processing_method",
        "cell_density_dilution_factor",
    ],
}

DATA_PROCESSING_FIELDS = [
    "cell_type_processing_method",
    "cell_density_dilution_factor",
]
